from __future__ import division

import numpy as np
from .utils import msqrt, check_type, round_array, float_dtypes, integer_dtypes, bool_dtypes, safe_len, find_generations, logp_of_set, symmetrize, logp_gradient_of_set, get_signature
from numpy import ndim, ones, zeros, log, shape, cov, ndarray, inner, reshape, sqrt, any, array, all, abs, exp, where, isscalar, iterable, multiply, transpose, tri, pi
from numpy.linalg.linalg import LinAlgError
from numpy.linalg import pinv, cholesky
from numpy.random import randint, random
from numpy.random import normal as rnormal
from numpy.random import poisson as rpoisson
from numpy.random import uniform as runiform
from numpy.random import exponential as rexponential
from .PyMCObjects import Stochastic, Potential, Deterministic
from .Container import Container
from .Node import ZeroProbability, Node, Variable, StochasticBase
from .decorators import prop
from . import distributions
from copy import copy
from .InstantiationDecorators import deterministic
import pdb
import warnings
import sys
import inspect

from . import six
from .six import print_

__docformat__ = 'reStructuredText'

conjugate_Gibbs_competence = 0
nonconjugate_Gibbs_competence = 0


class AdaptationError(ValueError):
    pass


__all__ = [
    'DiscreteMetropolis',
    'Metropolis',
    'PDMatrixMetropolis',
    'StepMethod',
    'assign_method',
    'pick_best_methods',
    'StepMethodRegistry',
    'NoStepper',
    'BinaryMetropolis',
    'AdaptiveMetropolis',
    'Gibbs',
    'conjugate_Gibbs_competence',
    'nonconjugate_Gibbs_competence',
    'DrawFromPrior',
    'Slicer']


StepMethodRegistry = []


def pick_best_methods(stochastic):
    """
    Picks the StepMethods best suited to handle
    a stochastic variable.
    """

    # Keep track of most competent methohd
    max_competence = 0
    # Empty set of appropriate StepMethods
    best_candidates = set([])

    # Loop over StepMethodRegistry
    for method in StepMethodRegistry:

        # Parse method and its associated competence
        try:
            competence = method.competence(stochastic)
        except:
            competence = 0

        # If better than current best method, promote it
        if competence > max_competence:
            best_candidates = set([method])
            max_competence = competence

        # If same competence, add it to the set of best methods
        elif competence == max_competence:
            best_candidates.add(method)

    if max_competence <= 0:
        raise ValueError(
            'Maximum competence reported for stochastic %s is <= 0... you may need to write a custom step method class.' %
            stochastic.__name__)

    # print_(s.__name__ + ': ', best_candidates, ' ', max_competence)
    return best_candidates


def assign_method(stochastic, scale=None, verbose=-1):
    """
    Returns a step method instance to handle a
    variable. If several methods have the same competence,
    it picks one arbitrarily (using set.pop()).
    """

    # Retrieve set of best candidates
    best_candidates = pick_best_methods(stochastic)

    # Randomly grab and appropriate method
    method = best_candidates.pop()

    failure_header = """Failed attempting to automatically assign step method class %s
to stochastic variable %s. Try setting %s's competence method to return 0
and manually assigning it when appropriate. See the user guide.

Error message: """ % (method.__name__, stochastic.__name__, method.__name__)

    try:
        if scale:
            out = method(stochastic, scale=scale, verbose=verbose)
        else:
            out = method(stochastic, verbose=verbose)
    except:
        a, b, c = sys.exc_info()
        try:
            args = list(b.args)
        except AttributeError:
            args = []
        args.append(failure_header)
        b.args = args
        six.reraise(a, b, c)
    return out


class StepMethodMeta(type):

    """
    Automatically registers new step methods if they can be automatically assigned:
    if their init method has one and only one required argument.
    """
    def __init__(cls, name, bases, dict):
        type.__init__(cls, name, bases, dict)
        args, defaults = get_signature(cls.__init__)
        auto_assignment_OK = False
        if len(args) == 2:
            auto_assignment_OK = True
        elif len(args) > 2:
            if defaults is not None:
                if len(defaults) == len(args) - 2:
                    auto_assignment_OK = True
        elif len(args) == 1 and varargs is not None:
            auto_assignment_OK = True

        if auto_assignment_OK:
            StepMethodRegistry.append(cls)


class StepMethod(object):

    """
    This object knows how to make Stochastics take single MCMC steps.
    Its step() method will be called by Model at every MCMC iteration.

    :Parameters:
          -variables : list, array or set
            Collection of PyMCObjects

          - verbose (optional) : integer
            Level of output verbosity: 0=none, 1=low, 2=medium, 3=high. Setting to -1 (Default) allows verbosity to be set by sampler.

    Externally-accessible attributes:
      stochastics:   The Stochastics over which self has jurisdiction which have observed = False.
      children: The combined children of all Variables over which self has jurisdiction.
      parents:  The combined parents of all Nodes over which self has jurisdiction, as a set.
      loglike:  The summed log-probability of self's children conditional on all of self's
                  Variables' current values. These will be recomputed only as necessary.
                  This descriptor should eventually be written in C.

    Externally accesible methods:
      sample(): A single MCMC step for all the Stochastics over which self has
        jurisdiction. Must be overridden in subclasses.
      tune(): Tunes proposal distribution widths for all self's Stochastics.
      competence(s): Examines Stochastic instance s and returns self's
        competence to handle it, on a scale of 0 to 3.

    To instantiate a StepMethod called S with jurisdiction over a
    sequence/set N of Nodes:

      >>> S = StepMethod(N)

    :SeeAlso: Metropolis, Sampler.
    """

    def __init__(self, variables, verbose=-1, tally=False):
        # StepMethod initialization

        if not iterable(variables) or isinstance(variables, Node):
            variables = [variables]

        self.stochastics = set()
        self.children = set()
        self.parents = set()
        self.tally = tally

        self._state = []
        self._tuning_info = []
        self.verbose = verbose

        # File away the variables
        for variable in variables:
            # Sort.

            if isinstance(variable, Stochastic) and not variable.observed:
                self.stochastics.add(variable)

        if len(self.stochastics) == 0:
            raise ValueError('No stochastics provided.')

        # Find children, no need to find parents; each variable takes care of
        # those.
        for variable in variables:
            self.children |= variable.children
            for parent in six.itervalues(variable.parents):
                if isinstance(parent, Variable):
                    self.parents.add(parent)

        self.children = set([])
        self.parents = set([])
        for s in self.stochastics:
            self.children |= s.extended_children
            self.parents |= s.extended_parents

        # Remove own stochastics from children and parents.
        self.children -= self.stochastics
        self.parents -= self.stochastics

        # self.markov_blanket is a list, because we want self.stochastics to have the chance to
        # raise ZeroProbability exceptions before self.children.
        self.markov_blanket = list(self.stochastics) + list(self.children)

        # ID string for verbose feedback
        self._id = self.__class__.__name__ + '_' + \
            '_'.join([s.__name__ for s in self.stochastics])

    def step(self):
        """
        Specifies single step of step method.
        Must be overridden in subclasses.
        """
        pass

    @staticmethod
    def competence(s):
        """
        This function is used by Sampler to determine which step method class
        should be used to handle stochastic variables.

        Return value should be a competence
        score from 0 to 3, assigned as follows:

        0:  I can't handle that variable.
        1:  I can handle that variable, but I'm a generalist and
            probably shouldn't be your top choice (Metropolis
            and friends fall into this category).
        2:  I'm designed for this type of situation, but I could be
            more specialized.
        3:  I was made for this situation, let me handle the variable.

        In order to be eligible for inclusion in the registry, a sampling
        method's init method must work with just a single argument, a
        Stochastic object.

        If you want to exclude a particular step method from
        consideration for handling a variable, do this:

        Competence functions MUST be called 'competence' and be decorated by the
        '@staticmethod' decorator. Example:

            @staticmethod
            def competence(s):
                if isinstance(s, MyStochasticSubclass):
                    return 2
                else:
                    return 0

        :SeeAlso: pick_best_methods, assign_method
        """
        return 0

    def tune(self, *args, **kwargs):
        return False

    @property
    def loglike(self):
        '''
        The summed log-probability of all stochastic variables that depend on
        self.stochastics, with self.stochastics removed.
        '''
        sum = logp_of_set(self.children)
        if self.verbose > 2:
            print_('\t' + self._id + ' Current log-likelihood ', sum)
        return sum

    @property
    def logp_plus_loglike(self):
        '''
        The summed log-probability of all stochastic variables that depend on
        self.stochastics, and self.stochastics.
        '''
        sum = logp_of_set(self.markov_blanket)
        if self.verbose > 2:
            print_('\t' + self._id +
                   ' Current log-likelihood plus current log-probability', sum)
        return sum

    @property
    def logp_gradient(self):
        return logp_gradient_of_set(self.stochastics, self.markov_blanket)

    def current_state(self):
        """Return a dictionary with the current value of the variables defining
        the state of the step method."""
        state = {}
        for s in self._state:
            state[s] = getattr(self, s)
        return state

    @prop
    def ratio():
        """Acceptance ratio"""

        def fget(self):
            return self.accepted / (self.accepted + self.rejected)
        return locals()

StepMethod = six.with_metaclass(StepMethodMeta, StepMethod)


class NoStepper(StepMethod):

    """
    Step and tune methods do nothing.

    Useful for holding stochastics constant without setting observed=True.
    """

    def step(self):
        pass

    def tune(self, *args, **kwargs):
        pass

# The default StepMethod, which Model uses to handle singleton stochastics.


class Metropolis(StepMethod):

    """
    The default StepMethod, which Model uses to handle singleton, continuous variables.

    Applies the one-at-a-time Metropolis-Hastings algorithm to the Stochastic over which self has jurisdiction.

    To instantiate a Metropolis called M with jurisdiction over a Stochastic P:

      >>> M = Metropolis(P, scale=1, proposal_sd=None, dist=None)

    :Arguments:
    - s : Stochastic
            The variable over which self has jurisdiction.

    - scale (optional) : number
            The proposal jump width is set to scale * variable.value.

    - proposal_sd (optional) : number or vector
            The proposal jump width is set to proposal_sd.

    - proposal_distribution (optional) : string
            The proposal distribution. May be 'normal' (default) or
            'prior'.

    - verbose (optional) : integer
            Level of output verbosity: 0=none, 1=low, 2=medium, 3=high. Setting to -1 (default) allows verbosity to be turned on by sampler.

    :SeeAlso: StepMethod, Sampler.
    """

    def __init__(self, stochastic, scale=1., proposal_sd=None,
                 proposal_distribution='normal', verbose=-1, tally=True, check_before_accepting=True):
        # Metropolis class initialization

        # Initialize superclass
        StepMethod.__init__(self, [stochastic], tally=tally)

        # Initialize hidden attributes
        self.proposal_sd = proposal_sd

        self.adaptive_scale_factor = 1.
        self.accepted = 0.
        self.rejected = 0.
        self._state = [
            'rejected',
            'accepted',
            'adaptive_scale_factor',
            'proposal_sd',
            'proposal_distribution',
            'check_before_accepting']
        self._tuning_info = ['adaptive_scale_factor']
        self.check_before_accepting = check_before_accepting
        self.proposal_sd = proposal_sd

        # Set public attributes
        self.stochastic = stochastic
        if verbose > -1:
            self.verbose = verbose
        else:
            self.verbose = stochastic.verbose
           
        prop_dist = proposal_distribution.lower() 
        if prop_dist in self.valid_proposals:
            self.proposal_distribution = proposal_distribution
        else:
            raise ValueError(
                "Invalid proposal distribution '%s' specified for Metropolis sampler." %
                proposal_distribution)

        if prop_dist != "prior":
            # Avoid zeros when setting proposal variance
            if proposal_sd is None:
                if all(self.stochastic.value != 0.):
                    self.proposal_sd = ones(
                        shape(
                            self.stochastic.value)) * abs(
                                self.stochastic.value) * scale
                else:
                    self.proposal_sd = ones(shape(
                        self.stochastic.value)) * scale

            # Initialize proposal deviate with array of zeros
            self.proposal_deviate = zeros(
                shape(self.stochastic.value),
                dtype=float)

            # Determine size of stochastic
            if isinstance(self.stochastic.value, ndarray):
                self._len = len(self.stochastic.value.ravel())
            else:
                self._len = 1

    valid_proposals = property(lambda self: ('normal', 'prior'))

    @staticmethod
    def competence(s):
        """
        The competence function for Metropolis
        """
        if s.dtype is None:
            return .5

        if not s.dtype in float_dtypes:
            # If the stochastic's binary or discrete, I can't do it.
            return 0
        else:
            return 2

    def hastings_factor(self):
        """
        If this is a Metropolis-Hastings method (proposal is not symmetric random walk),
        this method should return log(back_proposal) - log(forward_proposal).
        """
        return 0.

    def step(self):
        """
        The default step method applies if the variable is floating-point
        valued, and is not being proposed from its prior.
        """

        # Probability and likelihood for s's current value:

        if self.verbose > 2:
            print_()
            print_(self._id + ' getting initial logp.')

        if self.proposal_distribution == "prior":
            logp = self.loglike
        else:
            logp = self.logp_plus_loglike

        if self.verbose > 2:
            print_(self._id + ' proposing.')

        # Sample a candidate value
        self.propose()

        # Probability and likelihood for s's proposed value:
        try:
            if self.proposal_distribution == "prior":
                logp_p = self.loglike
                # Check for weirdness before accepting jump
                if self.check_before_accepting:
                    self.stochastic.logp
            else:
                logp_p = self.logp_plus_loglike

        except ZeroProbability:

            # Reject proposal
            if self.verbose > 2:
                print_(self._id + ' rejecting due to ZeroProbability.')
            self.reject()

            # Increment rejected count
            self.rejected += 1

            if self.verbose > 2:
                print_(self._id + ' returning.')
            return

        if self.verbose > 2:
            print_('logp_p - logp: ', logp_p - logp)

        HF = self.hastings_factor()

        # Evaluate acceptance ratio
        if log(random()) > logp_p - logp + HF:

            # Revert s if fail
            self.reject()

            # Increment rejected count
            self.rejected += 1
            if self.verbose > 2:
                print_(self._id + ' rejecting')
        else:
            # Increment accepted count
            self.accepted += 1
            if self.verbose > 2:
                print_(self._id + ' accepting')

        if self.verbose > 2:
            print_(self._id + ' returning.')

    def reject(self):
        # Sets current s value to the last accepted value
        # self.stochastic.value = self.stochastic.last_value
        self.stochastic.revert()

    def propose(self):
        """
        This method is called by step() to generate proposed values
        if self.proposal_distribution is "Normal" (i.e. no proposal specified).
        """
    
        prop_dist = self.proposal_distribution.lower()
        if  prop_dist == "normal":
            self.stochastic.value = rnormal(
                self.stochastic.value,
                self.adaptive_scale_factor *
                self.proposal_sd,
                size=self.stochastic.value.shape)
        elif prop_dist == "prior":
            self.stochastic.random()

    def tune(self, divergence_threshold=1e10, verbose=0):
        """
        Tunes the scaling parameter for the proposal distribution
        according to the acceptance rate of the last k proposals:

        Rate    Variance adaptation
        ----    -------------------
        <0.001        x 0.1
        <0.05         x 0.5
        <0.2          x 0.9
        >0.5          x 1.1
        >0.75         x 2
        >0.95         x 10

        This method is called exclusively during the burn-in period of the
        sampling algorithm.

        May be overridden in subclasses.
        """

        if self.verbose > -1:
            verbose = self.verbose

        # Verbose feedback
        if verbose > 0:
            print_('\t%s tuning:' % self._id)

        # Flag for tuning state
        tuning = True

        # Calculate recent acceptance rate
        if not (self.accepted + self.rejected):
            return tuning
        acc_rate = self.accepted / (self.accepted + self.rejected)
        
        current_factor = self.adaptive_scale_factor

        # Switch statement
        if acc_rate < 0.001:
            # reduce by 90 percent
            self.adaptive_scale_factor *= 0.1
        elif acc_rate < 0.05:
            # reduce by 50 percent
            self.adaptive_scale_factor *= 0.5
        elif acc_rate < 0.2:
            # reduce by ten percent
            self.adaptive_scale_factor *= 0.9
        elif acc_rate > 0.95:
            # increase by factor of ten
            self.adaptive_scale_factor *= 10.0
        elif acc_rate > 0.75:
            # increase by double
            self.adaptive_scale_factor *= 2.0
        elif acc_rate > 0.5:
            # increase by ten percent
            self.adaptive_scale_factor *= 1.1
        else:
            tuning = False

        # Re-initialize rejection count
        self.rejected = 0.
        self.accepted = 0.
        
        # Prevent from tuning to zero
        if not self.adaptive_scale_factor:
            self.adaptive_scale_factor = current_factor
            return False

        # More verbose feedback, if requested
        if verbose > 0:
            if hasattr(self, 'stochastic'):
                print_('\t\tvalue:', self.stochastic.value)
            print_('\t\tacceptance rate:', acc_rate)
            print_('\t\tadaptive scale factor:', self.adaptive_scale_factor)
            print_()

        return tuning


class PDMatrixMetropolis(Metropolis):

    """Metropolis sampler with proposals customised for symmetric positive definite matrices"""

    def __init__(self, stochastic, scale=1.,
                 proposal_sd=None, verbose=-1, tally=True):
        Metropolis.__init__(
            self,
            stochastic,
            scale=scale,
            proposal_sd=proposal_sd,
            proposal_distribution="normal",
            verbose=verbose,
            tally=tally)

    @staticmethod
    def competence(s):
        """
        The competence function for MatrixMetropolis
        """
        # MatrixMetropolis handles the Wishart family, which are valued as
        # _symmetric_ matrices.
        if any([isinstance(s, cls)
                for cls in [distributions.Wishart, distributions.WishartCov]]):
            return 2
        else:
            return 0

    def propose(self):
        """
        Proposals for positive definite matrix using random walk deviations on the Cholesky
        factor of the current value.
        """

        # Locally store size of matrix
        dims = self.stochastic.value.shape

        # Add normal deviate to value and symmetrize
        dev = rnormal(
            0,
            self.adaptive_scale_factor *
            self.proposal_sd,
            size=dims)
        symmetrize(dev)

        # Replace
        self.stochastic.value = dev + self.stochastic.value


class Gibbs(Metropolis):

    """
    Base class for the Gibbs step methods
    """

    def __init__(self, stochastic, verbose=-1):
        Metropolis.__init__(self, stochastic, verbose=verbose, tally=False)

    # Override Metropolis's competence.
    competence = classmethod(StepMethod.competence)

    def step(self):
        if not self.conjugate:
            logp = self.stochastic.logp

        self.propose()

        if not self.conjugate:

            try:
                logp_p = self.stochastic.logp
            except ZeroProbability:
                self.reject()

            if log(random()) > logp_p - logp:
                self.reject()

    def tune(self, *args, **kwargs):
        return False

    def propose(self):
        raise NotImplementedError(
            'The Gibbs class has to be subclassed, it is not usable directly.')


class DrawFromPrior(StepMethod):

    """
    Handles dataless submodels.
    """

    def __init__(self, variables, generations, verbose=-1):
        StepMethod.__init__(self, variables, verbose, tally=False)
        self.generations = generations

        # Some variables (eg GP) may not have logp attributes, so don't try to
        # evaluate their logps.
        self.variables_with_logp = set([])
        for s in self.markov_blanket:
            try:
                s.logp
                self.variables_with_logp.add(s)
            except:
                pass

    def get_logp_plus_loglike(self):
        return logp_of_set(self.variables_with_logp)
    logp_plus_loglike = property(get_logp_plus_loglike)

    def step(self):
        jumped = []
        try:
            for generation in self.generations:
                for s in generation:
                    s.rand()
                    jumped.append(s)
            self.logp_plus_loglike
        except ZeroProbability:
            if self.verbose > 2:
                forbidden = []
                for generation in self.generations:
                    for s in self.stochastics:
                        try:
                            s.logp
                        except ZeroProbability:
                            forbidden.append(s.__name__)
                print_(
                    'DrawFromPrior jumped stochastics %s to value forbidden by objects %s, rejecting.' % (
                        ', '.join(
                            s.__name__ for s in jumped),
                        ', '.join(
                            forbidden)))
            warnings.warn('DrawFromPrior jumped to forbidden value')
            for s in jumped:
                s.revert()

    @classmethod
    def competence(s):
        # Dataless gets assigned specially before other step methods.
        return 0


class DiscreteMetropolis(Metropolis):

    """
    Just like Metropolis, but rounds the variable's value.
    Good for discrete stochastics.
    """

    def __init__(self, stochastic, scale=1., proposal_sd=None,
                 proposal_distribution="poisson", non_negative=None, verbose=-1, tally=True):
        # DiscreteMetropolis class initialization

        # Initialize superclass
        Metropolis.__init__(
            self,
            stochastic,
            scale=scale,
            proposal_sd=proposal_sd,
            proposal_distribution=proposal_distribution,
            verbose=verbose,
            tally=tally)
            
        if non_negative is None:
            
            init_val = stochastic.value
            if np.any(init_val <= 0):
                non_negative = False
                
            else:
                # If -1 has finite logp, allow positive 
                stochastic.value = -np.ones_like(init_val)
                try:
                    non_negative = ~np.isfinite(stochastic.logp)
                except ZeroProbability:
                    non_negative = True
                    
                stochastic.value = init_val
        
        else:
            non_negative = True
        
        # Flag for positive-only values
        self._non_negative = non_negative
        
    valid_proposals = property(lambda self: ('normal', 'prior', 'poisson'))

    @staticmethod
    def competence(stochastic):
        """
        The competence function for DiscreteMetropolis.
        """
        if stochastic.dtype in integer_dtypes:
            return 1
        else:
            return 0

    def propose(self):
        # Propose new values using normal distribution

        prop_dist = self.proposal_distribution.lower()
        if prop_dist == "normal":

            # New normal deviate, centred on current value
            new_val = rnormal(
                self.stochastic.value,
                self.adaptive_scale_factor *
                self.proposal_sd)

            # Round before setting proposed value
            self.stochastic.value = round_array(new_val)

        elif prop_dist == "poisson":

            k = shape(self.stochastic.value)
            # Add or subtract (equal probability) Poisson sample
            new_val = self.stochastic.value + rpoisson(
                self.adaptive_scale_factor * self.proposal_sd) * (
                    -ones(k)) ** (random(k) > 0.5)

            if self._non_negative:
                # Enforce positive values
                self.stochastic.value = abs(new_val)
            else:
                self.stochastic.value = new_val

        elif prop_dist == "prior":
            self.stochastic.random()

# TODO Implement independence sampler for BinaryMetropolis


class BinaryMetropolis(Metropolis):

    """
    Like Metropolis, but with a modified step() method.
    Good for binary variables.

    """

    def __init__(self, stochastic, p_jump=.1,
                 proposal_distribution='binary', verbose=-1, tally=True):
        # BinaryMetropolis class initialization

        # Initialize superclass
        Metropolis.__init__(
            self,
            stochastic,
            proposal_distribution=proposal_distribution,
            verbose=verbose,
            tally=tally)

        self._state.remove('proposal_sd')

        # adaptive_scale_factor controls the jump probability
        self.adaptive_scale_factor = log(1. - p_jump) / log(.5)
        
    valid_proposals = property(lambda self: ('binary', 'prior'))

    @staticmethod
    def competence(stochastic):
        """
        The competence function for Binary One-At-A-Time Metropolis
        """
        if stochastic.dtype in bool_dtypes:
            return 2

        elif isinstance(stochastic, distributions.Bernoulli):
            return 2
            
        elif (isinstance(stochastic, distributions.Categorical) and 
            (len(stochastic.parents['p'])==2)):
            return 2

        else:
            return 0

    def step(self):
        if ndim(self.stochastic.value):
            Metropolis.step(self)
        else:

            # See what log-probability of True is.
            self.stochastic.value = True

            try:
                logp_true = self.logp_plus_loglike
            except ZeroProbability:
                self.stochastic.value = False
                return

            # See what log-probability of False is.
            self.stochastic.value = False

            try:
                logp_false = self.logp_plus_loglike
            except ZeroProbability:
                self.stochastic.value = True
                return

            # Test
            p_true = exp(logp_true)
            p_false = exp(logp_false)

            if self.verbose > 2:
                print_("""%s step information:
                - logp_true: %f
                - logp_false: %f
                - p_true: %f
                - p_false: %f
                """ % (self._id, logp_true, logp_false, p_true, p_false))

            # Stochastically set value according to relative
            # probabilities of True and False
            if random() > p_false / (p_true + p_false):
                if self.verbose > 2:
                    print_(
                        "%s setting %s's value to True." %
                        (self._id, self.stochastic))
                self.stochastic.value = True
            elif self.verbose > 2:
                print_(
                    "%s setting %s's value to False." %
                    (self._id, self.stochastic))

    def propose(self):

        if self.proposal_distribution.lower() == 'prior':
            self.stochastic.random()
        else:
            # Convert adaptive_scale_factor to a jump probability
            p_jump = 1. - .5 ** self.adaptive_scale_factor

            rand_array = random(size=shape(self.stochastic.value))
            new_value = copy(self.stochastic.value)
            # Locations where switches occur, according to p_jump
            switch_locs = where(rand_array < p_jump)
            if shape(new_value):
                new_value[switch_locs] = True - new_value[switch_locs]
            else:
                new_value = True - new_value
            self.stochastic.value = new_value


class AdaptiveMetropolis(StepMethod):

    """
    The AdaptativeMetropolis (AM) sampling algorithm works like a regular
    Metropolis, with the exception that stochastic parameters are block-updated
    using a multivariate jump distribution whose covariance is tuned during
    sampling. Although the chain is non-Markovian, i.e. the proposal
    distribution is asymmetric, it has correct ergodic properties. See
    (Haario et al., 2001) for details.

    :Parameters:
      - stochastic : PyMC objects
          Stochastic objects to be handled by the AM algorith,

      - cov : array
          Initial guess for the covariance matrix C. If it is None, the
          covariance will be estimated using the scales dictionary if provided,
          the existing trace if available, or the current stochastics value.
          It is suggested to provide a sensible guess for the covariance, and
          not rely on the automatic assignment from stochastics value.

      - delay : int
          Number of steps before the empirical covariance is computed. If greedy
          is True, the algorithm waits for delay *accepted* steps before computing
          the covariance.

      - interval : int
          Interval between covariance updates. Higher dimensional spaces require
          more samples to obtain reliable estimates for the covariance updates.

      - greedy : bool
          If True, only the accepted jumps are tallied in the internal trace
          until delay is reached. This is useful to make sure that the empirical
          covariance has a sensible structure.

      - shrink_if_necessary : bool
          If True, the acceptance rate is checked when the step method tunes. If
          the acceptance rate is small, the proposal covariance is shrunk according
          to the following rule:

          if acc_rate < .001:
              self.C *= .01
          elif acc_rate < .01:
              self.C *= .25

      - scales : dict
          Dictionary containing the scale for each stochastic keyed by name.
          If cov is None, those scales are used to define an initial covariance
          matrix. If neither cov nor scale is given, the initial covariance is
          guessed from the trace (it if exists) or the objects value, alt

      - verbose : int
          Controls the verbosity level.


    :Notes:
    Use the methods: `cov_from_scales`, `cov_from_trace` and `cov_from_values` for
    more control on the creation of an initial covariance matrix. A lot of problems
    can be avoided with a good initial covariance and long enough intervals between
    covariance updates. That is, do not compensate for a bad covariance guess by
    reducing the interval between updates thinking the covariance matrix will
    converge more rapidly.


    :Reference:
      Haario, H., E. Saksman and J. Tamminen, An adaptive Metropolis algorithm,
          Bernouilli, vol. 7 (2), pp. 223-242, 2001.
    """

    def __init__(self, stochastic, cov=None, delay=1000, interval=200,
                 greedy=True, shrink_if_necessary=False, scales=None, verbose=-1, tally=False):

        # Verbosity flag
        self.verbose = verbose

        self.accepted = 0
        self.rejected = 0

        if not np.iterable(stochastic) or isinstance(stochastic, Variable):
            stochastic = [stochastic]

        # Initialize superclass
        StepMethod.__init__(self, stochastic, verbose, tally)

        self._id = 'AdaptiveMetropolis_' + '_'.join(
            [p.__name__ for p in self.stochastics])
        # State variables used to restore the state in a latter session.
        self._state += [
            'accepted', 'rejected', '_trace_count', '_current_iter', 'C', 'proposal_sd',
            '_proposal_deviate', '_trace', 'shrink_if_necessary']
        self._tuning_info = ['C']

        self.proposal_sd = None
        self.shrink_if_necessary = shrink_if_necessary

        # Number of successful steps before the empirical covariance is
        # computed
        self.delay = delay
        # Interval between covariance updates
        self.interval = interval
        # Flag for tallying only accepted jumps until delay reached
        self.greedy = greedy

        # Initialization methods
        self.check_type()
        self.dimension()

        # Set the initial covariance using cov, or the following fallback mechanisms:
        # 1. If scales is provided, use it.
        # 2. If a trace is present, compute the covariance matrix empirically from it.
        # 3. Use the stochastics value as a guess of the variance.
        if cov is not None:
            self.C = cov
        elif scales:
            self.C = self.cov_from_scales(scales)
        else:
            try:
                self.C = self.cov_from_trace()
            except AttributeError:
                self.C = self.cov_from_value(100.)

        self.updateproposal_sd()

        # Keep track of the internal trace length
        # It may be different from the iteration count since greedy
        # sampling can be done during warm-up period.
        self._trace_count = 0
        self._current_iter = 0

        self._proposal_deviate = np.zeros(self.dim)
        self.chain_mean = np.asmatrix(np.zeros(self.dim))
        self._trace = []

        if self.verbose >= 2:
            print_("Initialization...")
            print_('Dimension: ', self.dim)
            print_("C_0: ", self.C)
            print_("Sigma: ", self.proposal_sd)

    @staticmethod
    def competence(stochastic):
        """
        The competence function for AdaptiveMetropolis.
        The AM algorithm is well suited to deal with multivariate
        parameters, particularly those which are correlated with one
        another. However, it does not work reliably with all multivariate
        stochastics, so it must be applied manually via MCMC.use_step_method().
        """
        return 0

    def cov_from_value(self, scaling):
        """Return a covariance matrix for the jump distribution using
        the actual value of the stochastic as a guess of their variance,
        divided by the `scaling` argument.

        Note that this is likely to return a poor guess.
        """
        rv = []
        for s in self.stochastics:
            rv.extend(np.ravel(s.value).copy())

        # Remove 0 values since this would lead to quite small jumps...
        arv = np.array(rv)
        arv[arv == 0] = 1.

        # Create a diagonal covariance matrix using the scaling factor.
        return np.eye(self.dim) * np.abs(arv) / scaling

    def cov_from_scales(self, scales):
        """Return a covariance matrix built from a dictionary of scales.

        `scales` is a dictionary keyed by stochastic instances, and the
        values refer are the variance of the jump distribution for each
        stochastic. If a stochastic is a sequence, the variance must
        have the same length.
        """

        # Get array of scales
        ord_sc = []
        for stochastic in self.stochastics:
            ord_sc.append(np.ravel(scales[stochastic]))
        ord_sc = np.concatenate(ord_sc)

        if np.squeeze(ord_sc).shape[0] != self.dim:
            raise ValueError("Improper initial scales, dimension don't match",
                             (np.squeeze(ord_sc), self.dim))

        # Scale identity matrix
        return np.eye(self.dim) * ord_sc

    def cov_from_trace(self, trace=slice(None)):
        """Define the jump distribution covariance matrix from the object's
        stored trace.

        :Parameters:
        - `trace` : slice or int
          A slice for the stochastic object's trace in the last chain, or a
          an integer indicating the how many of the last samples will be used.

        """
        n = []
        for s in self.stochastics:
            n.append(s.trace.length())
        n = set(n)
        if len(n) > 1:
            raise ValueError('Traces do not have the same length.')
        elif n == 0:
            raise AttributeError(
                'Stochastic has no trace to compute covariance.')
        else:
            n = n.pop()

        if not isinstance(trace, slice):
            trace = slice(trace, n)

        a = self.trace2array(trace)

        return np.cov(a, rowvar=0)

    def check_type(self):
        """Make sure each stochastic has a correct type, and identify discrete stochastics."""
        self.isdiscrete = {}
        for stochastic in self.stochastics:
            if stochastic.dtype in integer_dtypes:
                self.isdiscrete[stochastic] = True
            elif stochastic.dtype in bool_dtypes:
                raise ValueError(
                    'Binary stochastics not supported by AdaptativeMetropolis.')
            else:
                self.isdiscrete[stochastic] = False

    def dimension(self):
        """Compute the dimension of the sampling space and identify the slices
        belonging to each stochastic.
        """
        self.dim = 0
        self._slices = {}
        for stochastic in self.stochastics:
            if isinstance(stochastic.value, np.matrix):
                p_len = len(stochastic.value.A.ravel())
            elif isinstance(stochastic.value, np.ndarray):
                p_len = len(stochastic.value.ravel())
            else:
                p_len = 1
            self._slices[stochastic] = slice(self.dim, self.dim + p_len)
            self.dim += p_len

    def update_cov(self):
        """Recursively compute the covariance matrix for the multivariate normal
        proposal distribution.

        This method is called every self.interval once self.delay iterations
        have been performed.
        """

        scaling = (2.4) ** 2 / self.dim  # Gelman et al. 1996.
        epsilon = 1.0e-5
        chain = np.asarray(self._trace)

        # Recursively compute the chain mean
        self.C, self.chain_mean = self.recursive_cov(self.C, self._trace_count,
                                                     self.chain_mean, chain, scaling=scaling, epsilon=epsilon)

        # Shrink covariance if acceptance rate is too small
        acc_rate = self.accepted / (self.accepted + self.rejected)
        if self.shrink_if_necessary:
            if acc_rate < .001:
                self.C *= .01
            elif acc_rate < .01:
                self.C *= .25
            if self.verbose > 1:
                if acc_rate < .01:
                    print_(
                        '\tAcceptance rate was',
                        acc_rate,
                        'shrinking covariance')
        self.accepted = 0.
        self.rejected = 0.

        if self.verbose > 1:
            print_("\tUpdating covariance ...\n", self.C)
            print_("\tUpdating mean ... ", self.chain_mean)

        # Update state
        adjustmentwarning = '\n' +\
            'Covariance was not positive definite and proposal_sd cannot be computed by \n' + \
            'Cholesky decomposition. The next jumps will be based on the last \n' + \
            'valid covariance matrix. This situation may have arisen because no \n' + \
            'jumps were accepted during the last `interval`. One solution is to \n' + \
            'increase the interval, or specify an initial covariance matrix with \n' + \
            'a smaller variance. For this simulation, each time a similar error \n' + \
            'occurs, proposal_sd will be reduced by a factor .9 to reduce the \n' + \
            'jumps and increase the likelihood of accepted jumps.'

        try:
            self.updateproposal_sd()
        except np.linalg.LinAlgError:
            warnings.warn(adjustmentwarning)
            self.covariance_adjustment(.9)

        self._trace_count += len(self._trace)
        self._trace = []

    def covariance_adjustment(self, f=.9):
        """Multiply self.proposal_sd by a factor f. This is useful when the current proposal_sd is too large and all jumps are rejected.
        """
        self.proposal_sd *= f

    def updateproposal_sd(self):
        """Compute the Cholesky decomposition of self.C."""
        self.proposal_sd = np.linalg.cholesky(self.C)

    def recursive_cov(self, cov, length, mean, chain, scaling=1, epsilon=0):
        r"""Compute the covariance recursively.

        Return the new covariance and the new mean.

        .. math::
            C_k & = \frac{1}{k-1} (\sum_{i=1}^k x_i x_i^T - k\bar{x_k}\bar{x_k}^T)
            C_n & = \frac{1}{n-1} (\sum_{i=1}^k x_i x_i^T + \sum_{i=k+1}^n x_i x_i^T - n\bar{x_n}\bar{x_n}^T)
                & = \frac{1}{n-1} ((k-1)C_k + k\bar{x_k}\bar{x_k}^T + \sum_{i=k+1}^n x_i x_i^T - n\bar{x_n}\bar{x_n}^T)

        :Parameters:
            -  cov : matrix
                Previous covariance matrix.
            -  length : int
                Length of chain used to compute the previous covariance.
            -  mean : array
                Previous mean.
            -  chain : array
                Sample used to update covariance.
            -  scaling : float
                Scaling parameter
            -  epsilon : float
                Set to a small value to avoid singular matrices.
        """
        n = length + len(chain)
        k = length
        new_mean = self.recursive_mean(mean, length, chain)

        t0 = k * np.outer(mean, mean)
        t1 = np.dot(chain.T, chain)
        t2 = n * np.outer(new_mean, new_mean)
        t3 = epsilon * np.eye(cov.shape[0])

        new_cov = (
            k - 1) / (
                n - 1.) * cov + scaling / (
                    n - 1.) * (
                        t0 + t1 - t2 + t3)
        return new_cov, new_mean

    def recursive_mean(self, mean, length, chain):
        r"""Compute the chain mean recursively.

        Instead of computing the mean :math:`\bar{x_n}` of the entire chain,
        use the last computed mean :math:`bar{x_j}` and the tail of the chain
        to recursively estimate the mean.

        .. math::
            \bar{x_n} & = \frac{1}{n} \sum_{i=1}^n x_i
                      & = \frac{1}{n} (\sum_{i=1}^j x_i + \sum_{i=j+1}^n x_i)
                      & = \frac{j\bar{x_j}}{n} + \frac{\sum_{i=j+1}^n x_i}{n}

        :Parameters:
            -  mean : array
                Previous mean.
            -  length : int
                Length of chain used to compute the previous mean.
            -  chain : array
                Sample used to update mean.
        """
        n = length + len(chain)
        return length * mean / n + chain.sum(0) / n

    def propose(self):
        """
        This method proposes values for stochastics based on the empirical
        covariance of the values sampled so far.

        The proposal jumps are drawn from a multivariate normal distribution.
        """

        arrayjump = np.dot(
            self.proposal_sd,
            np.random.normal(
                size=self.proposal_sd.shape[
                    0]))
        if self.verbose > 2:
            print_('Jump :', arrayjump)

        # Update each stochastic individually.
        for stochastic in self.stochastics:
            jump = arrayjump[self._slices[stochastic]].squeeze()
            if np.iterable(stochastic.value):
                jump = np.reshape(
                    arrayjump[
                        self._slices[
                            stochastic]],
                    np.shape(
                        stochastic.value))
            if self.isdiscrete[stochastic]:
                jump = round_array(jump)
            stochastic.value = stochastic.value + jump

    def step(self):
        """
        Perform a Metropolis step.

        Stochastic parameters are block-updated using a multivariate normal
        distribution whose covariance is updated every self.interval once
        self.delay steps have been performed.

        The AM instance keeps a local copy of the stochastic parameter's trace.
        This trace is used to computed the empirical covariance, and is
        completely independent from the Database backend.

        If self.greedy is True and the number of iterations is smaller than
        self.delay, only accepted jumps are stored in the internal
        trace to avoid computing singular covariance matrices.
        """

        # Probability and likelihood for stochastic's current value:
        logp = self.logp_plus_loglike
        if self.verbose > 1:
            print_('Current value: ', self.stoch2array())
            print_('Current likelihood: ', logp)

        # Sample a candidate value
        self.propose()

        # Metropolis acception/rejection test
        accept = False
        try:
            # Probability and likelihood for stochastic's proposed value:
            logp_p = self.logp_plus_loglike
            if self.verbose > 2:
                print_('Current value: ', self.stoch2array())
                print_('Current likelihood: ', logp_p)

            if np.log(random()) < logp_p - logp:
                accept = True
                self.accepted += 1
                if self.verbose > 2:
                    print_('Accepted')
            else:
                self.rejected += 1
                if self.verbose > 2:
                    print_('Rejected')
        except ZeroProbability:
            self.rejected += 1
            logp_p = None
            if self.verbose > 2:
                print_('Rejected with ZeroProbability Error.')

        if (not self._current_iter % self.interval) and self.verbose > 1:
            print_("Step ", self._current_iter)
            print_("\tLogprobability (current, proposed): ", logp, logp_p)
            for stochastic in self.stochastics:
                print_(
                    "\t",
                    stochastic.__name__,
                    stochastic.last_value,
                    stochastic.value)
            if accept:
                print_("\tAccepted\t*******\n")
            else:
                print_("\tRejected\n")
            print_(
                "\tAcceptance ratio: ",
                self.accepted / (
                    self.accepted + self.rejected))

        if self._current_iter == self.delay:
            self.greedy = False

        if not accept:
            self.reject()

        if accept or not self.greedy:
            self.internal_tally()

        if self._current_iter > self.delay and self._current_iter % self.interval == 0:
            self.update_cov()

        self._current_iter += 1

    # Please keep reject() factored out- helps RandomRealizations figure out
    # what to do.
    def reject(self):
        for stochastic in self.stochastics:
            # stochastic.value = stochastic.last_value
            stochastic.revert()

    def internal_tally(self):
        """Store the trace of stochastics for the computation of the covariance.
        This trace is completely independent from the backend used by the
        sampler to store the samples."""
        chain = []
        for stochastic in self.stochastics:
            chain.append(np.ravel(stochastic.value))
        self._trace.append(np.concatenate(chain))

    def trace2array(self, sl):
        """Return an array with the trace of all stochastics, sliced by sl."""
        chain = []
        for stochastic in self.stochastics:
            tr = stochastic.trace.gettrace(slicing=sl)
            if tr is None:
                raise AttributeError
            chain.append(tr)
        return np.hstack(chain)

    def stoch2array(self):
        """Return the stochastic objects as an array."""
        a = np.empty(self.dim)
        for stochastic in self.stochastics:
            a[self._slices[stochastic]] = stochastic.value
        return a

    def tune(self, verbose=0):
        """Tuning is done during the entire run, independently from the Sampler
        tuning specifications. """
        return False


class TWalk(StepMethod):

    """
    The t-walk is a scale-independent, adaptive MCMC algorithm for arbitrary
    continuous distributions and correltation structures. The t-walk maintains
    two independent points in the sample space, and moves are based on
    proposals that are accepted or rejected with a standard M-H acceptance
    probability on the product space. The t-walk is strictly non-adaptive on
    the product space, but displays adaptive behaviour on the original state
    space. There are four proposal distributions (walk, blow, hop, traverse)
    that together offer an algorithm that is effective in sampling
    distributions of arbitrary scale.

    The t-walk was devised by J.A. Christen and C. Fox (2010).

    :Parameters:
      - stochastic : Stochastic
          The variable over which self has jurisdiction.
      - kernel_probs (optional) : iterable
          The probabilities of choosing each kernel.
      - walk_theta (optional) : float
          Parameter for the walk move. Christen and Fox recommend
          values in [0.3, 2] (Defaults to 1.5).
      - traverse_theta (optional) : float
          Parameter for the traverse move. Christen and Fox recommend
          values in [2, 10] (Defaults to 6.0).
      - n1 (optional) : integer
          The number of elements to be moved at each iteration.
          Christen and Fox recommend values in [2, 20] (Defaults to 4).
      - support (optional) : function
          Function defining the support of the stochastic
          (Defaults to real line).
      - verbose (optional) : integer
          Level of output verbosity: 0=none, 1=low, 2=medium, 3=high
      - tally (optional) : bool
          Flag for recording values for trace (Defaults to True).
    """

    def __init__(self, stochastic, inits=None, kernel_probs=[
                 0.4918, 0.4918, 0.0082, 0.0082], walk_theta=1.5, traverse_theta=6.0, n1=4, support=lambda x: True, verbose=-1, tally=True):

        # Initialize superclass
        StepMethod.__init__(self, [stochastic], verbose=verbose, tally=tally)

        # Ordered list of proposal kernels
        self.kernels = [self.walk, self.traverse, self.blow, self.hop]

        # Kernel for current iteration
        self.current_kernel = None

        self.accepted = zeros(len(kernel_probs))
        self.rejected = zeros(len(kernel_probs))

        # Cumulative kernel probabilities
        self.cum_probs = np.cumsum(kernel_probs)

        self.walk_theta = walk_theta
        self.traverse_theta = traverse_theta

        # Set public attributes
        self.stochastic = stochastic
        if verbose > -1:
            self.verbose = verbose
        else:
            self.verbose = stochastic.verbose

        # Determine size of stochastic
        if isinstance(self.stochastic.value, ndarray):
            self._len = len(self.stochastic.value.ravel())
        else:
            self._len = 1

        # Create attribute for holding value and secondary value
        self.values = [self.stochastic.value]

        # Initialize to different value from stochastic or supplied values
        if inits is None:
            self.values.append(self.stochastic.random())
            # Reset original value
            self.stochastic.value = self.values[0]
        else:
            self.values.append(inits)

        # Flag for using second point in log-likelihood calculations
        self._prime = False

        # Proposal adjustment factor for current iteration
        self.hastings_factor = 0.0

        # Set probability of selecting any parameter
        self.p = 1. * min(self._len, n1) / self._len

        # Support function
        self._support = support

        self._state = ['accepted', 'rejected', 'p']

    def n1():
        doc = "Mean number of parameters to be selected for updating"

        def fget(self):
            return self._n1

        def fset(self, value):
            self._n1 = value
            self._calc_p()
        return locals()
    n1 = property(**n1())

    @staticmethod
    def competence(stochastic):
        """
        The competence function for TWalk.
        """
        # if stochastic.dtype in float_dtypes and np.alen(stochastic.value) > 4:
        #             if np.alen(stochastic.value) >=10:
        #                 return 2
        #             return 1
        return 0

    def walk(self):
        """Walk proposal kernel"""

        if self.verbose > 1:
            print_('\t' + self._id + ' Running Walk proposal kernel')

        # Mask for values to move
        phi = self.phi

        theta = self.walk_theta

        u = random(len(phi))
        z = (theta / (1 + theta)) * (theta * u ** 2 + 2 * u - 1)

        if self._prime:
            xp, x = self.values
        else:
            x, xp = self.values

        if self.verbose > 1:
            print_('\t' + 'Current value = ' + str(x))

        x = x + phi * (x - xp) * z

        if self.verbose > 1:
            print_('\t' + 'Proposed value = ' + str(x))

        self.stochastic.value = x

        # Set proposal adjustment factor
        self.hastings_factor = 0.0

    def traverse(self):
        """Traverse proposal kernel"""

        if self.verbose > 1:
            print_('\t' + self._id + ' Running Traverse proposal kernel')

        # Mask for values to move
        phi = self.phi

        theta = self.traverse_theta

        # Calculate beta
        if (random() < (theta - 1) / (2 * theta)):
            beta = exp(1 / (theta + 1) * log(random()))
        else:
            beta = exp(1 / (1 - theta) * log(random()))

        if self._prime:
            xp, x = self.values
        else:
            x, xp = self.values

        if self.verbose > 1:
            print_('\t' + 'Current value = ' + str(x))

        x = (xp + beta * (xp - x)) * phi + x * (phi == False)

        if self.verbose > 1:
            print_('\t' + 'Proposed value = ' + str(x))

        self.stochastic.value = x

        # Set proposal adjustment factor
        self.hastings_factor = (sum(phi) - 2) * log(beta)

    def blow(self):
        """Blow proposal kernel"""

        if self.verbose > 1:
            print_('\t' + self._id + ' Running Blow proposal kernel')

        # Mask for values to move
        phi = self.phi

        if self._prime:
            xp, x = self.values
        else:
            x, xp = self.values

        if self.verbose > 1:
            print_('\t' + 'Current value ' + str(x))

        sigma = max(phi * abs(xp - x))

        x = x + phi * sigma * rnormal()

        if self.verbose > 1:
            print_('\t' + 'Proposed value = ' + str(x))

        self.hastings_factor = self._g(
            x,
            xp,
            sigma) - self._g(
                self.stochastic.value,
                xp,
                sigma)

        self.stochastic.value = x

    def _g(self, h, xp, s):
        """Density function for blow and hop moves"""

        nphi = sum(self.phi)

        return (nphi / 2.0) * log(2 * pi) + nphi * \
            log(s) + 0.5 * sum((h - xp) ** 2) / (s ** 2)

    def hop(self):
        """Hop proposal kernel"""

        if self.verbose > 1:
            print_('\t' + self._id + ' Running Hop proposal kernel')

        # Mask for values to move
        phi = self.phi

        if self._prime:
            xp, x = self.values
        else:
            x, xp = self.values

        if self.verbose > 1:
            print_('\t' + 'Current value of x = ' + str(x))

        sigma = max(phi * abs(xp - x)) / 3.0

        x = (xp + sigma * rnormal()) * phi + x * (phi == False)

        if self.verbose > 1:
            print_('\t' + 'Proposed value = ' + str(x))

        self.hastings_factor = self._g(
            x,
            xp,
            sigma) - self._g(
                self.stochastic.value,
                xp,
                sigma)

        self.stochastic.value = x

    def reject(self):
        """Sets current s value to the last accepted value"""
        self.stochastic.revert()

        # Increment rejected count
        self.rejected[self.current_kernel] += 1

        if self.verbose > 1:
            print_(
                self._id,
                "rejected, reverting to value =",
                self.stochastic.value)

    def propose(self):
        """This method is called by step() to generate proposed values"""

        # Generate uniform variate to choose kernel
        self.current_kernel = sum(self.cum_probs < random())
        kernel = self.kernels[self.current_kernel]

        # Parameters to move
        self.phi = (random(self._len) < self.p)

        # Propose new value
        kernel()

    def step(self):
        """Single iteration of t-walk algorithm"""

        valid_proposal = False

        # Use x or xprime as pivot
        self._prime = (random() < 0.5)

        if self.verbose > 1:
            print_("\n\nUsing x%s as pivot" % (" prime" * self._prime or ""))

        if self._prime:
            # Set the value of the stochastic to the auxiliary
            self.stochastic.value = self.values[1]

            if self.verbose > 1:
                print_(
                    self._id,
                    "setting value to auxiliary",
                    self.stochastic.value)

        # Current log-probability
        logp = self.logp_plus_loglike
        if self.verbose > 1:
            print_("Current logp", logp)

        try:
            # Propose new value
            while not valid_proposal:
                self.propose()
                # Check that proposed value lies in support
                valid_proposal = self._support(self.stochastic.value)

            if not sum(self.phi):
                raise ZeroProbability

            # Proposed log-probability
            logp_p = self.logp_plus_loglike
            if self.verbose > 1:
                print_("Proposed logp", logp_p)

        except ZeroProbability:

            # Reject proposal
            if self.verbose > 1:
                print_(self._id + ' rejecting due to ZeroProbability.')
            self.reject()

            if self._prime:
                # Update value list
                self.values[1] = self.stochastic.value
                # Revert to stochastic's value for next iteration
                self.stochastic.value = self.values[0]

                if self.verbose > 1:
                    print_(
                        self._id,
                        "reverting stochastic to primary value",
                        self.stochastic.value)
            else:
                # Update value list
                self.values[0] = self.stochastic.value

            if self.verbose > 1:
                print_(self._id + ' returning.')
            return

        if self.verbose > 1:
            print_('logp_p - logp: ', logp_p - logp)

        # Evaluate acceptance ratio
        if log(random()) > (logp_p - logp + self.hastings_factor):

            # Revert s if fail
            self.reject()

        else:
            # Increment accepted count
            self.accepted[self.current_kernel] += 1
            if self.verbose > 1:
                print_(self._id + ' accepting')

        if self._prime:
            # Update value list
            self.values[1] = self.stochastic.value
            # Revert to stochastic's value for next iteration
            self.stochastic.value = self.values[0]

            if self.verbose > 1:
                print_(
                    self._id,
                    "reverting stochastic to primary value",
                    self.stochastic.value)

        else:
            # Update value list
            self.values[0] = self.stochastic.value

# Slice sampler implementation contributed by Dominik Wabersich


class Slicer(StepMethod):

    """
    Univariate slice sampler step method

    :Parameters:
      - stochastic : Stochastic
          The variable over which self has jurisdiction.
      - w (optional): float
          Initial width of slice (Defaults to 1)
      - m (optional): int
          Multiplier defining maximum slice size to :math:`mw` (Defaults to 1000)
      - tune (optional): bool
          Tune initial slice width (defaults to True)
      - doubling (optional): bool
          Flag for using doubling procedure instead of stepping out (Defaults to False)
      - tally (optional) : bool
          Flag for recording values for trace (Defaults to True).
      - verbose(optional) : int
          Set verbosity level (Defaults to -1)
    """

    def __init__(self, stochastic, w=1, m=1000, tune=True,
                 doubling=False, tally=False, verbose=-1):
        """
        Slice sampler class initialization
        """
        # Initialize superclass
        StepMethod.__init__(self, [stochastic], tally=tally)

        # id string
        self._id = "Slicer"

        # Set public attributes
        self.stochastic = stochastic
        if verbose > -1:
            self.verbose = verbose
        else:
            self.verbose = stochastic.verbose

        self._tune = tune
        if tune:
            self.w_tune = []
        self.w = w
        self.m = m
        self.doubling = doubling

    @staticmethod
    def competence(s):
        """
        The competence function for Slice

        Works best for continuous scalar variables.
        """
        if (s.dtype in float_dtypes) and not s.shape:
            return 1
        else:
            return 0

    def step(self):
        """
        Slice step method

        From Neal 2003 (doi:10.1214/aos/1056562461)
        """
        logy = self.loglike - rexponential(1)

        L = self.stochastic.value - runiform(0, self.w)
        R = L + self.w

        if self.doubling:
            # Doubling procedure
            K = self.m
            while (K and (logy < self.fll(L) or logy < self.fll(R))):
                if random() < 0.5:
                    L -= R - L
                else:
                    R += R - L
                K -= 1
        else:
            # Stepping out procedure
            J = np.floor(runiform(0, self.m))
            K = (self.m - 1) - J
            while(J > 0 and logy < self.fll(L)):
                L -= self.w
                J -= 1
            while(K > 0 and logy < self.fll(R)):
                R += self.w
                K -= 1

        # Shrinkage procedure
        self.stochastic.value = runiform(L, R)
        try:
            logy_new = self.loglike
        except ZeroProbability:
            logy_new = -np.infty
        while(logy_new < logy):
            if (self.stochastic.value < self.stochastic.last_value):
                L = float(self.stochastic.value)
            else:
                R = float(self.stochastic.value)
            self.stochastic.revert()
            self.stochastic.value = runiform(L, R)
            try:
                logy_new = self.loglike
            except ZeroProbability:
                logy_new = -np.infty

    def fll(self, value):
        """
        Returns loglike of value
        """
        self.stochastic.value = value
        try:
            ll = self.loglike
        except ZeroProbability:
            ll = -np.infty
        self.stochastic.revert()
        return ll

    def tune(self, verbose=None):
        """
        Tuning initial slice width parameter
        """
        if not self._tune:
            return False
        else:
            self.w_tune.append(
                abs(self.stochastic.last_value - self.stochastic.value))
            self.w = 2 * (sum(self.w_tune) / len(self.w_tune))
            return True



from numpy import *
from pymc import *
from numpy.testing import *
import nose
import sys
from pymc import utils
from pymc import six
import pymc

random.seed(21421)

float_dtypes = [float, single, float_, longfloat]


def generate_model():

    # Simulated data
    ReData = arange(200, 3000, 25)
    measured = 10.2 * (
        ReData) ** .5 + random.normal(
            scale=55,
            size=size(
                ReData))

    sd = pymc.Uniform('sd', lower=5, upper=100)

    a = pymc.Uniform('a', lower=0, upper=100)
    b = pymc.Uniform('b', lower=.05, upper=2.0)

    nonlinear = a * (ReData) ** b
    precision = sd ** -2

    results = pymc.Normal(
        'results',
        mu=nonlinear,
        tau=precision,
        value=measured,
        observed=True)

    return locals()


def find_variable_set(stochastic):
    set = [stochastic]
    for parameter, variable in six.iteritems(stochastic.parents):
        if isinstance(variable, Variable):
            set.append(variable)
    return set


def check_jacobians(deterministic):
    for parameter, pvalue in six.iteritems(deterministic.parents):

        if isinstance(pvalue, Variable):

            grad = random.normal(.5, .1, size=shape(deterministic.value))
            a_partial_grad = get_analytic_partial_gradient(
                deterministic, parameter, pvalue, grad)

            n_partial_grad = get_numeric_partial_gradient(
                deterministic, pvalue, grad)

            assert_array_almost_equal(a_partial_grad, n_partial_grad, 4,
                                      "analytic partial gradient for " + str(deterministic) +
                                      " with respect to parameter " + str(parameter) +
                                      " is not correct.")


def get_analytic_partial_gradient(deterministic, parameter, variable, grad):
    try:
        jacobian = deterministic._jacobians[parameter].get()
    except KeyError:
        raise ValueError(str(
            deterministic) + " has no jacobian for " + str(parameter))
    mapping = deterministic._jacobian_formats.get(parameter, 'full')

    return deterministic._format_mapping[mapping](
        deterministic, variable, jacobian, grad)


def get_numeric_partial_gradient(deterministic, pvalue, grad):
    j = get_numeric_jacobian(deterministic, pvalue)
    pg = deterministic._format_mapping['full'](deterministic, pvalue, j, grad)
    return reshape(pg, shape(pvalue.value))


def get_numeric_jacobian(deterministic, pvalue):
    e = 1e-9
    initial_pvalue = pvalue.value
    shape = initial_pvalue.shape
    size = initial_pvalue.size

    initial_value = ravel(deterministic.value)

    numeric_jacobian = zeros((deterministic.value.size, size))
    for i in range(size):

        delta = zeros(size)
        delta[i] += e

        pvalue.value = reshape(initial_pvalue.ravel() + delta, shape)
        value = ravel(deterministic.value)

        numeric_jacobian[:, i] = (value - initial_value) / e

    pvalue.value = initial_pvalue
    return numeric_jacobian


def check_model_gradients(model):

    model = set(model)
    # find the markov blanket
    children = set([])
    for s in model:
        for s2 in s.extended_children:
            if isinstance(s2, Stochastic) and s2.observed:
                children.add(s2)

    # self.markov_blanket is a list, because we want self.stochastics to have the chance to
    # raise ZeroProbability exceptions before self.children.
    markov_blanket = list(model) + list(children)

    gradients = utils.logp_gradient_of_set(model)
    for variable in model:

        analytic_gradient = gradients[variable]

        numeric_gradient = get_numeric_gradient(markov_blanket, variable)

        assert_array_almost_equal(numeric_gradient, analytic_gradient, 3,
                                  "analytic gradient for model " + str(model) +
                                  " with respect to variable " + str(variable) +
                                  " is not correct.")


def check_gradients(stochastic):

    stochastics = find_variable_set(stochastic)
    gradients = utils.logp_gradient_of_set(stochastics, stochastics)

    for s, analytic_gradient in six.iteritems(gradients):

        numeric_gradient = get_numeric_gradient(stochastics, s)

        assert_array_almost_equal(numeric_gradient, analytic_gradient, 3,
                                  "analytic gradient for " + str(stochastic) +
                                  " with respect to parameter " + str(s) +
                                  " is not correct.")


def get_numeric_gradient(stochastic, pvalue):
    e = 1e-9
    initial_value = pvalue.value
    i_shape = shape(initial_value)
    i_size = size(initial_value)

    initial_logp = utils.logp_of_set(stochastic)
    numeric_gradient = zeros(i_shape)
    if not (pvalue.dtype in float_dtypes):
        return numeric_gradient

    for i in range(i_size):

        delta = zeros(i_shape)

        try:
            delta[unravel_index(i, i_shape or (1,))] += e
        except IndexError:
            delta += e

        pvalue.value = initial_value + delta
        logp = utils.logp_of_set(stochastic)

        try:
            numeric_gradient[
                unravel_index(
                    i,
                    i_shape or (
                        1,
                    ))] = (
                        logp - initial_logp) / e
        except IndexError:
            numeric_gradient = (logp - initial_logp) / e

    pvalue.value = initial_value
    return numeric_gradient


class test_gradients(TestCase):

    def test_jacobians(self):
        shape = (3, 10)
        a = Normal('a', mu=zeros(shape), tau=ones(shape))
        b = Normal('b', mu=zeros(shape), tau=ones(shape))
        c = Uniform('c', lower=ones(shape) * .1, upper=ones(shape) * 10)
        d = Uniform('d', lower=ones(shape) * -10, upper=ones(shape) * -.1)

        addition = a + b
        check_jacobians(addition)

        subtraction = a - b
        check_jacobians(subtraction)

        multiplication = a * b
        check_jacobians(subtraction)

        division1 = a / c
        check_jacobians(division1)

        division2 = a / d
        check_jacobians(division2)

        a2 = Uniform('a2', lower=.1 * ones(shape), upper=2.0 * ones(shape))
        powering = a2 ** b
        check_jacobians(powering)

        negation = -a
        check_jacobians(negation)

        absing = abs(a)
        check_jacobians(absing)

        indexing1 = a[0:1, 5:8]
        check_jacobians(indexing1)

        indexing3 = a[::-1, :]
        check_jacobians(indexing3)

        # this currently does not work because scalars use the Index deterministic
        # which is special and needs more thought
        indexing2 = a[0]
        check_jacobians(indexing2)

    def test_numpy_deterministics_jacobians(self):

        shape = (2, 3)
        a = Normal('a', mu=zeros(shape), tau=ones(shape) * 5)
        b = Normal('b', mu=zeros(shape), tau=ones(shape))
        c = Uniform('c', lower=ones(shape) * .1, upper=ones(shape) * 10)
        d = Uniform('d', lower=ones(shape) * -1.0, upper=ones(shape) * 1.0)
        e = Normal('e', mu=zeros(shape), tau=ones(shape))
        f = Uniform('c', lower=ones(shape) * 1.0, upper=ones(shape) * 10)

        summing = sum(a, axis=0)
        check_jacobians(summing)

        summing2 = sum(a)
        check_jacobians(summing2)

        absing = abs(a)
        check_jacobians(absing)

        exping = exp(a)
        check_jacobians(exping)

        logging = log(c)
        check_jacobians(logging)

        sqrting = sqrt(c)
        check_jacobians(sqrting)

        sining = sin(a)
        check_jacobians(sining)

        cosing = cos(a)
        check_jacobians(cosing)

        taning = tan(a)
        check_jacobians(taning)

        arcsining = arcsin(d)
        check_jacobians(arcsining)

        arcosing = arccos(d)
        check_jacobians(arcosing)

        arctaning = arctan(d)
        check_jacobians(arctaning)

        sinhing = sinh(a)
        check_jacobians(sinhing)

        coshing = cosh(a)
        check_jacobians(coshing)

        tanhing = tanh(a)
        check_jacobians(tanhing)

        arcsinhing = arcsinh(a)
        check_jacobians(arcsinhing)

        arccoshing = arccosh(f)
        check_jacobians(arccoshing)

        arctanhing = arctanh(d)
        check_jacobians(arctanhing)

        arctan2ing = arctan2(b, e)
        check_jacobians(arctan2ing)

        hypoting = hypot(b, e)
        check_jacobians(hypoting)

    def test_gradients(self):

        shape = (5,)
        a = Normal('a', mu=zeros(shape), tau=ones(shape))
        b = Normal('b', mu=zeros(shape), tau=ones(shape))
        b2 = Normal('b2', mu=2, tau=1.0)
        c = Uniform('c', lower=ones(shape) * .7, upper=ones(shape) * 2.5)
        d = Uniform('d', lower=ones(shape) * .7, upper=ones(shape) * 2.5)
        e = Uniform('e', lower=ones(shape) * .2, upper=ones(shape) * 10)
        f = Uniform('f', lower=ones(shape) * 2, upper=ones(shape) * 30)
        p = Uniform(
            'p',
            lower=zeros(
                shape) + .05,
            upper=ones(
                shape) - .05)
        n = 5

        a.value = 2 * ones(shape)
        b.value = 2.5 * ones(shape)
        b2.value = 2.5

        norm = Normal('norm', mu=a, tau=b)
        check_gradients(norm)

        norm2 = Normal('norm2', mu=0, tau=b2)
        check_gradients(norm2)

        gamma = Gamma('gamma', alpha=a, beta=b)
        check_gradients(gamma)

        bern = Bernoulli('bern', p=p)
        check_gradients(bern)

        beta = Beta('beta', alpha=c, beta=d)

        check_gradients(beta)

        cauchy = Cauchy('cauchy', alpha=a, beta=d)
        check_gradients(cauchy)

        chi2 = Chi2('chi2', nu=e)
        check_gradients(chi2)

        exponential = Exponential('expon', beta=d)
        check_gradients(exponential)

        t = T('t', nu=f)
        check_gradients(t)

        half_normal = HalfNormal('half_normal', tau=e)
        check_gradients(half_normal)

        inverse_gamma = InverseGamma('inverse_gamma', alpha=c, beta=d)
        check_gradients(inverse_gamma)

        laplace = Laplace('laplace', mu=a, tau=c)
        check_gradients(laplace)

        lognormal = Lognormal('lognormal', mu=a, tau=c)
        check_gradients(lognormal)

        weibull = Weibull('weibull', alpha=c, beta=d)
        check_gradients(weibull)

        binomial = Binomial('binomial', p=p, n=n)
        check_gradients(binomial)

        geometric = Geometric('geometric', p=p)
        check_gradients(geometric)

        poisson = Poisson('poisson', mu=c)
        check_gradients(poisson)

        u = Uniform('u', lower=a, upper=b)
        check_gradients(u)

        negative_binomial = NegativeBinomial(
            'negative_binomial',
            mu=c,
            alpha=d)
        check_gradients(negative_binomial)

        # exponweib still not working for unknown reasons
        # exponweib = Exponweib('exponweib', alpha = c, k =d , loc = a, scale = e )
        # check_gradients(exponweib)

    def test_model(self):

        model = generate_model()
        model["sd"].value = 55.0
        model["a"].value = 10.2
        model["b"].value = .5

        check_gradients(model["sd"])

        M = pymc.MCMC(model)
        check_model_gradients(M.stochastics)


if __name__ == '__main__':
    C = nose.config.Config(verbosity=1)
    nose.runmodule(config=C)



# Need to accomodate for SciPy's recent changes in organization
# of statistical distributuions
try:
    from scipy.stats import _continuous_distns as cdists
    from scipy.stats import _discrete_distns as ddists
    rv_discrete = ddists.rv_discrete
    rv_continuous = cdists.rv_continuous
    sp_old = False
except ImportError:
    import scipy.stats.distributions as sc_dst
    rv_discrete = sc_dst.rv_discrete
    rv_continuous = sc_dst.rv_continuous
    sp_old = True
import inspect
import numpy as np
from . import Stochastic
from copy import copy
from .distributions import *
from .utils import get_signature

"""
Wraps a SciPy rv object in a PyMC random variable object.
Needs to wrap the following methods:

generic.rvs(<shape(s)>,loc=0,scale=1)
    - random variates

generic.pdf(x,<shape(s)>,loc=0,scale=1)
    - probability density function

generic.cdf(x,<shape(s)>,loc=0,scale=1)
    - cumulative density function

generic.sf(x,<shape(s)>,loc=0,scale=1)
    - survival function (1-cdf --- sometimes more accurate)

generic.ppf(q,<shape(s)>,loc=0,scale=1)
    - percent point function (inverse of cdf --- percentiles)

generic.isf(q,<shape(s)>,loc=0,scale=1)
    - inverse survival function (inverse of sf)

generic.stats(<shape(s)>,loc=0,scale=1,moments='mv')
    - mean('m',axis=0), variance('v'), skew('s'), and/or kurtosis('k')

generic.entropy(<shape(s)>,loc=0,scale=1)
    - (differential) entropy of the RV.
"""
__all__ = ['stochastic_from_scipy_dist']


def separate_shape_args(kwds, shape_args):

    new_kwds = {}
    for name in kwds:
        new_kwds[name] = kwds[name]

    args = [new_kwds.pop(name) for name in shape_args]

    return args, new_kwds


def remove_size_arg(fun):
    def wrapper(size=1, *args, **kwds):
        return fun(*args, **kwds)


def stochastic_from_scipy_dist(scipy_dist):
    """
    Return a Stochastic subclass made from a particular SciPy distribution.
    """

    name = scipy_dist.__class__.__name__.replace('_gen', '').capitalize()

    (args, defaults) = get_signature(scipy_dist._cdf)
    shape_args = args[2:]
    if isinstance(scipy_dist, rv_continuous):
        dtype = float

        def logp(value, **kwds):
            args, kwds = separate_shape_args(kwds, shape_args)
            return np.sum(scipy_dist.logpdf(value, *args, **kwds))

        parent_names = shape_args + ['loc', 'scale']
        defaults = [None] * (len(parent_names) - 2) + [0., 1.]

    elif isinstance(scipy_dist, rv_discrete):
        dtype = int

        def logp(value, **kwds):
            args, kwds = separate_shape_args(kwds, shape_args)
            return np.sum(scipy_dist.logpmf(value, *args, **kwds))

        parent_names = shape_args + ['loc']
        defaults = [None] * (len(parent_names) - 1) + [0]
    else:
        return None

    parents_default = dict(zip(parent_names, defaults))

    def random(shape=None, **kwds):
        args, kwds = separate_shape_args(kwds, shape_args)

        if shape is None:
            return scipy_dist.rvs(*args, **kwds)
        else:
            return np.reshape(scipy_dist.rvs(*args, **kwds), shape)

    # Build docstring from distribution
    docstr = name[0] + ' = ' + name + '(name, ' + ', '.join(
        parent_names) + ', value=None, shape=None, trace=True, rseed=True, doc=None)\n\n'
    docstr += 'Stochastic variable with ' + name + \
        ' distribution.\nParents are: ' + ', '.join(parent_names) + '.\n\n'
    docstr += """
Methods:

    random()
        - draws random value
          sets value to return value

    ppf(q)
        - percent point function (inverse of cdf --- percentiles)
          sets value to return value

    isf(q)
        - inverse survival function (inverse of sf)
          sets value to return value

    stats(moments='mv')
        - mean('m',axis=0), variance('v'), skew('s'), and/or kurtosis('k')


Attributes:

    logp
        - sum(log(pdf())) or sum(log(pmf()))

    cdf
        - cumulative distribution function

    sf
        - survival function (1-cdf --- sometimes more accurate)

    entropy
        - (differential) entropy of the RV.


NOTE: If you encounter difficulties with this object, please try the analogous
computation using the rv objects in scipy.stats.distributions directly before
reporting the bug.
    """

    new_class = new_dist_class(
        dtype,
        name,
        parent_names,
        parents_default,
        docstr,
        logp,
        random,
        True,
        None)

    class newer_class(new_class):
        __doc__ = docstr
        rv = scipy_dist

        def __init__(self, *args, **kwds):
            new_class.__init__(self, *args, **kwds)
            self.args, self.kwds = separate_shape_args(
                self.parents, shape_args)
            self.frozen_rv = self.rv(self.args, self.kwds)

        def _pymc_dists_to_value(self, args):
            """Replace arguments that are a pymc.Node with their value."""
            # This is needed because the scipy rv function transforms
            # every input argument which causes new pymc lambda
            # functions to be generated. Thus, when calling this many
            # many times, excessive amounts of RAM are used.
            new_args = []
            for arg in args:
                if isinstance(arg, pm.Node):
                    new_args.append(arg.value)
                else:
                    new_args.append(arg)
            return new_args

        def _cdf(self):
            """
            The cumulative distribution function of self conditional on parents
            evaluated at self's current value
            """
            return self.rv.cdf(
                self.value, *self._pymc_dists_to_value(self.args), **self.kwds)
        cdf = property(_cdf, doc=_cdf.__doc__)

        def _sf(self):
            """
            The survival function of self conditional on parents
            evaluated at self's current value
            """
            return self.rv.sf(
                self.value, *self._pymc_dists_to_value(self.args), **self.kwds)
        sf = property(_sf, doc=_sf.__doc__)

        def ppf(self, q):
            """
            The percentile point function (inverse cdf) of self conditional on parents.
            Self's value will be set to the return value.
            """
            self.value = self.rv.ppf(
                q,
                *
                self._pymc_dists_to_value(
                    self.args),
                **self.kwds)
            return self.value

        def isf(self, q):
            """
            The inverse survival function of self conditional on parents.
            Self's value will be set to the return value.
            """
            self.value = self.rv.isf(
                q,
                *self._pymc_dists_to_value(
                    self.args),
                **self.kwds)
            return self.value

        def stats(self, moments='mv'):
            """The first few moments of self's distribution conditional on parents"""
            return self.rv.stats(
                moments=moments, *self._pymc_dists_to_value(self.args), **self.kwds)

        def _entropy(self):
            """The entropy of self's distribution conditional on its parents"""
            return self.rv.entropy(
                *self._pymc_dists_to_value(self.args), **self.kwds)
        entropy = property(_entropy, doc=_entropy.__doc__)

    newer_class.__name__ = new_class.__name__
    return newer_class

if sp_old:

    for scipy_dist_name in sc_dst.__all__:
        scipy_dist = sc_dst.__dict__[scipy_dist_name]
        if isinstance(scipy_dist, rv_continuous) or isinstance(
                scipy_dist, rv_discrete):
            new_dist = stochastic_from_scipy_dist(scipy_dist)
            locals()[new_dist.__name__] = new_dist
            __all__.append(new_dist.__name__)

else:

    for scipy_dist_name in ddists.__all__:
        scipy_dist = ddists.__dict__[scipy_dist_name]
        if isinstance(scipy_dist, rv_discrete):
            new_dist = stochastic_from_scipy_dist(scipy_dist)
            locals()[new_dist.__name__] = new_dist
            __all__.append(new_dist.__name__)

    for scipy_dist_name in cdists.__all__:
        scipy_dist = cdists.__dict__[scipy_dist_name]
        if isinstance(scipy_dist, rv_continuous):
            new_dist = stochastic_from_scipy_dist(scipy_dist)
            locals()[new_dist.__name__] = new_dist
            __all__.append(new_dist.__name__)



# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'
__all__ = ['observe', 'plot_envelope', 'predictive_check', 'regularize_array', 'trimult', 'trisolve', 'vecs_to_datmesh', 'caching_call', 'caching_callable',
           'fast_matrix_copy', 'point_predict', 'square_and_sum', 'point_eval']


# TODO: Implement lintrans, allow obs_V to be a huge matrix or an ndarray
# in observe().

from numpy import *
from numpy.linalg import solve, cholesky, eigh
from numpy.linalg.linalg import LinAlgError
from .linalg_utils import *
from threading import Thread, Lock
import sys
from pymc import thread_partition_array, map_noreturn
from pymc.gp import chunksize
import pymc

from pymc import six
from pymc.six import print_
xrange = six.moves.xrange

try:
    from PyMC2 import ZeroProbability
except ImportError:
    class ZeroProbability(ValueError):
        pass

half_log_2pi = .5 * log(2. * pi)


def fast_matrix_copy(f, t=None, n_threads=1):
    """
    Not any faster than a serial copy so far.
    """
    if not f.flags['F_CONTIGUOUS']:
        raise RuntimeError(
            'This will not be fast unless input array f is Fortran-contiguous.')

    if t is None:
        t = asmatrix(empty(f.shape, order='F'))
    elif not t.flags['F_CONTIGUOUS']:
        raise RuntimeError(
            'This will not be fast unless input array t is Fortran-contiguous.')

    # Figure out how to divide job up between threads.
    dcopy_wrap(ravel(asarray(f.T)), ravel(asarray(t.T)))
    return t


def zero_lower_triangle(C):
    pass


def caching_call(f, x, x_sofar, f_sofar):
    """
    Computes f(x) given that f(x_sofar) = x_sofar.
    returns f(x), and new versions of x_sofar and f_sofar.
    """
    lenx = x.shape[0]

    nr, rf, rt, nu, xu, ui = remove_duplicates(x)

    unique_indices = ui[:nu]
    x_unique = xu[:nu]
    repeat_from = rf[:nr]
    repeat_to = rt[:nr]

    # Check which observations have already been made.
    if x_sofar is not None:

        f_unique, new_indices, N_new_indices = check_repeats(
            x_unique, x_sofar, f_sofar)

        # If there are any new input points, draw values over them.
        if N_new_indices > 0:

            x_new = x_unique[new_indices[:N_new_indices]]
            f_new = f(x_new)

            f_unique[new_indices[:N_new_indices]] = f_new

            # Record the new values
            x_sofar = vstack((x_sofar, x_new))
            f_sofar = hstack((f_sofar, f_new))
        else:
            f = f_unique

    # If no observations have been made, don't check.
    else:
        f_unique = f(x_unique)
        x_sofar = x_unique
        f_sofar = f_unique

    f = empty(lenx)
    f[unique_indices] = f_unique
    f[repeat_to] = f[repeat_from]

    return f, x_sofar, f_sofar


def square_and_sum(a, s):
    """
    Writes np.sum(a**2,axis=0) into s
    """
    cmin, cmax = thread_partition_array(a)
    map_noreturn(asqs, [(a, s, cmin[i], cmax[i]) for i in xrange(len(cmax))])
    return a


class caching_callable(object):

    """
    F = caching_callable(f[, x_sofar, f_sofar, update_cache=True])

    f is the function whose output should be cached.
    x_sofar, if provided, is an initial list of caching locations.
    f_sofar, if provided, is the value of f at x_sofar.
    update_cache tells whether x_sofar and f_sofar should be updated as additional calls are made.
    """

    def __init__(self, f, x_sofar=None, f_sofar=None, update_cache=True):
        self.f = f
        self.x_sofar = x_sofar
        self.f_sofar = f_sofar
        self.update_cache = update_cache
        if self.x_sofar is not None and self.f_sofar is None:
            junk, self.x_sofar, self.f_sofar = caching_call(
                self.f, self.x_sofar)
        self.last_x = x_sofar
        self.last_f = f_sofar

    def __call__(self, x):
        if x is self.x_sofar:
            return self.f_sofar
        elif x is self.last_x:
            return self.last_f
        f, x_sofar, f_sofar = caching_call(
            self.f, x, self.x_sofar, self.f_sofar)
        self.last_x = x
        self.last_f = f
        if self.update_cache:
            self.x_sofar = x_sofar
            self.f_sofar = f_sofar
        return f


def vecs_to_datmesh(x, y):
    """
    Converts input arguments x and y to a 2d meshgrid,
    suitable for calling Means, Covariances and Realizations.
    """
    x, y = meshgrid(x, y)
    out = zeros(x.shape + (2,), dtype=float)
    out[:, :, 0] = x
    out[:, :, 1] = y
    return out


def trimult(U, x, uplo='U', transa='N', alpha=1., inplace=False):
    """
    b = trimult(U,x, uplo='U')


    Multiplies U x, where U is upper triangular if uplo='U'
    or lower triangular if uplo = 'L'.
    """
    if inplace:
        b = x
    else:
        b = x.copy('F')
    dtrmm_wrap(a=U, b=b, uplo=uplo, transa=transa, alpha=alpha)
    return b


def trisolve(U, b, uplo='U', transa='N', alpha=1., inplace=False):
    """
    x = trisolve(U,b, uplo='U')


    Solves U x = b, where U is upper triangular if uplo='U'
    or lower triangular if uplo = 'L'.


    If a degenerate column is found, an error is raised.
    """
    if inplace:
        x = b
    else:
        x = b.copy('F')
    if U.shape[0] == 0:
        raise ValueError('Attempted to solve zero-rank triangular system')
    dtrsm_wrap(a=U, b=x, side='L', uplo=uplo, transa=transa, alpha=alpha)
    return x


def regularize_array(A):
    """
    Takes an ndarray as an input.


    -   If the array is one-dimensional, it's assumed to be an array of input values.

    -   If the array is more than one-dimensional, its last index is assumed to curse
        over spatial dimension.


    Either way, the return value is at least two dimensional. A.shape[-1] gives the
    number of spatial dimensions.
    """
    # Make sure A is an array.
    if not isinstance(A, ndarray):
        A = array(A, dtype=float)
    elif A.__class__ is not ndarray:
        A = asarray(A, dtype=float)

    # If A is one-dimensional, interpret it as an array of points on the line.
    if len(A.shape) <= 1:
        return A.reshape(-1, 1)

    # Otherwise, interpret it as an array of n-dimensional points, where n
    # is the size of A along its last index.
    elif A.shape[-1] > 1 and len(A.shape) > 2:
        return A.reshape(-1, A.shape[-1])

    else:
        return A


def plot_envelope(M, C, mesh):
    """
    plot_envelope(M,C,mesh)


    plots the pointwise mean +/- sd envelope defined by M and C
    along their base mesh.


    :Arguments:

        -   `M`: A Gaussian process mean.

        -   `C`: A Gaussian process covariance

        -   `mesh`: The mesh on which to evaluate the mean and cov.
    """

    try:
        from pylab import fill, plot, clf, axis
        x = concatenate((mesh, mesh[::-1]))
        mean, var = point_eval(M, C, mesh)
        sig = sqrt(var)
        mean = M(mesh)
        y = concatenate((mean - sig, (mean + sig)[::-1]))
        # clf()
        fill(x, y, facecolor='.8', edgecolor='1.')
        plot(mesh, mean, 'k-.')
    except ImportError:
        print_("Matplotlib is not installed; plotting is disabled.")


def observe(M, C, obs_mesh, obs_vals, obs_V=0, lintrans=None,
            cross_validate=True):
    """
    (M, C, obs_mesh, obs_vals[, obs_V = 0, lintrans = None, cross_validate = True])


    Imposes observation of the value of obs_vals on M and C, where

    obs_vals ~ N(lintrans * f(obs_mesh), V)
    f ~ GP(M,C)


    :Arguments:

        -   `M`: The mean function

        -   `C`: The covariance function

        -   `obs_mesh`: The places where f has been evaluated.

        -   `obs_vals`: The values of f that were observed there.

        -   `obs_V`: The observation variance. If None, assumed to be infinite
            (observations made with no error).

        -   `lintrans`: A linear transformation. If None, assumed to be the
            identity transformation (pretend it doesn't exist).

        -   `cross_validate`: A flag indicating whether a check should be done to
            see if the data could have arisen from M and C with positive probability.

    """
    obs_mesh = regularize_array(obs_mesh)
    # print_(obs_mesh)
    obs_V = resize(obs_V, obs_mesh.shape[0])
    obs_vals = resize(obs_vals, obs_mesh.shape[0])

    # First observe C.
    relevant_slice, obs_mesh_new = C.observe(obs_mesh, obs_V, output_type='o')

    # Then observe M from C.
    M.observe(C, obs_mesh_new, obs_vals.ravel()[relevant_slice])

    # Cross-validate if not asked not to.
    if obs_mesh_new.shape[0] < obs_mesh.shape[0]:
        if cross_validate:
            if not predictive_check(
                    obs_vals, obs_mesh, M, C.obs_piv, sqrt(C.relative_precision)):
                raise ValueError(
                    "These data seem extremely improbable given your GP prior. \n Suggestions: decrease observation precision, or adjust the covariance to \n allow the function to be less smooth.")


def predictive_check(obs_vals, obs_mesh, M, posdef_indices, tolerance):
    """
    OK = predictive_check(obs_vals, obs_mesh, M, posdef_indices, tolerance)


    If an internal covariance is low-rank, make sure the observations
    are consistent. Returns True if good, False if bad.


    :Arguments:

        -   `obs_vals`: The observed values.

        -   `obs_mesh`: The mesh on which the observed values were observed.

        -   `M`: The mean function, observed at obs_vals[posdef_indices].

        -   `tolerance`: The maximum allowable deviation at M(obs_mesh[non_posdef_indices]).
    """

    non_posdef_indices = array(
        list(set(range(len(obs_vals))) - set(posdef_indices)), dtype=int)
    if len(non_posdef_indices) > 0:
        M_under = M(obs_mesh[non_posdef_indices, :]).ravel()
        dev = abs((M_under - obs_vals[non_posdef_indices]))
        if dev.max() > tolerance:
            return False

    return True


def point_predict(f, x, size=1, nugget=None):
    """
    point_predict(f, x[, size, nugget])

    Makes 'size' simulations for f(x) + N(0,nugget).
    Simulated values of f(x_i) are uncorrelated for different i.
    Useful for geostatistical predictions.
    """
    orig_shape = x.shape
    x = regularize_array(x)

    mu = f.M_internal(x, regularize=False)
    V = f.C_internal(x, regularize=False)
    if nugget is not None:
        V += nugget
    out = random.normal(size=(size, x.shape[0])) * sqrt(V) + mu
    return out.reshape((size,) + orig_shape[:-1]).squeeze()


def point_eval(M, C, x):
    """
    Evaluates M(x) and C(x).

    Minimizes computation; evaluating M(x) and C(x) separately would
    evaluate the off-diagonal covariance term twice, but callling
    point_eval(M,C,x) would only evaluate it once.

    Also chunks the evaluations if the off-diagonal term.
    """

    x_ = regularize_array(x)

    M_out = empty(x_.shape[0])
    V_out = empty(x_.shape[0])

    if isinstance(C, pymc.gp.BasisCovariance):
        y_size = len(C.basis)
    elif C.obs_mesh is not None:
        y_size = C.obs_mesh.shape[0]
    else:
        y_size = 1

    n_chunks = ceil(y_size * x_.shape[0] / float(chunksize))
    bounds = array(linspace(0, x_.shape[0], n_chunks + 1), dtype='int')
    cmin = bounds[:-1]
    cmax = bounds[1:]
    for (cmin, cmax) in zip(bounds[:-1], bounds[1:]):
        x__ = x_[cmin:cmax]
        V_out[cmin:cmax], Uo_Cxo = C(x__, regularize=False, return_Uo_Cxo=True)
        M_out[cmin:cmax] = M(x__, regularize=False, Uo_Cxo=Uo_Cxo)

    if len(x.shape) > 1:
        targ_shape = x.shape[:-1]
    else:
        targ_shape = x.shape
    return M_out.reshape(targ_shape), V_out.reshape(targ_shape)



from .PyMCObjects import Stochastic
from .Container import Container
from .InstantiationDecorators import stochastic
from .flib import mod_to_circle
from .distributions import rvon_mises, von_mises_like
import numpy as np
from .distributions import valuewrapper

__all__ = ['CircularStochastic', 'CircVonMises']


class CircularStochastic(Stochastic):

    """
    C = CircularStochastic(lo, hi, *args, **kwargs)

    Takes two special parents, lo and hi; any incoming value
    will be mapped into the interval [lo,hi).

    args and kwargs will be passed to Stochastic.__init__.

    :Attributes:
      lo : float
        Lower bound
      hi : float
        Upper bound

    :SeeAlso: Stochastic
    """

    def __init__(self, lo, hi, *args, **kwargs):
        self.interval_parents = Container([hi, lo])
        Stochastic.__init__(self, *args, **kwargs)

    def set_value(self, value):
        modded_value = mod_to_circle(
            value,
            *
            self.interval_parents.value).reshape(
                np.shape(
                    value))
        Stochastic.set_value(self, modded_value)
    value = property(Stochastic.get_value, set_value)


class CircVonMises(CircularStochastic):

    """
    V = CircVonMises(name, mu, kappa, value=None, observed=False, size=1, trace=True, rseed=True, doc=None)

    Stochastic variable with Von Mises distribution.
    Parents are: mu, kappa.

    Docstring of log-probability function:
    """ + von_mises_like.__doc__

    def __init__(self,
                 name,
                 mu, kappa,
                 value=None,
                 observed=False,
                 size=1,
                 trace=True,
                 cache_depth=2,
                 rseed=True,
                 plot=None,
                 verbose=-1):

        if value is None:
            arg_eval = Container([mu, kappa]).value
            value = rvon_mises(arg_eval[0], arg_eval[1], size=size)

        parents = {'mu': mu, 'kappa': kappa}
        logp = valuewrapper(von_mises_like)
        random = lambda mu, kappa, size=size: rvon_mises(mu, kappa, size=size)
        CircularStochastic.__init__(
            self,
            -np.pi,
            np.pi,
            logp,
            'A Von Mises-distributed variable',
            name,
            parents,
            random,
            trace,
            value,
            np.dtype('float'),
            rseed,
            observed,
            cache_depth,
            plot,
            verbose)



"""
Base backend

Trace and Database classes from the other modules should Subclass the base
classes.

Concepts
--------

Each backend must define a Trace and a Database class that subclass the
base.Trace and base.Database classes.

The Model itself is given a `db` attribute, an instance of the Database
class. The __init__ method of the Database class does the work of preparing the
database. This might mean creating a file, connecting to a remote database
server, etc.

Right after a Database is instantiated, the Sampler tells it to connect itself
the model. This is taken care of by the connect_model method. This method
creates Trace objects for all the tallyable pymc objects. These traces are
stored in a dictionary called _traces, owned by the Database instance of this
Sampler. Previously, the Trace instances were owned by the tallyable variables
themselves. As A.P. pointed out, this is problematic if the same object is used
in different models, because each new Sampler will overwrite the last Trace
instance of the variable.

When the `sample` method is called, the Sampler tells the Database to initialize
itself. Database then tells all the tallyable objects to initialize themselves.
This might mean creating a numpy array, a list, an sqlite table, etc.

When the Sampler calls `tally`, it sends a message to the Database to tally its
objects. Each object then appends its current value to the given chain.

At the end of sampling, the Sampler tells the Database to finalize it's state,
and the database relays the finalize call to each Trace object.

To get the samples from the database, the Sampler class provides a trace method
``trace(self, name, chain=-1)`` which first sets the db attribute
_default_chain to chain and returns the trace instance, so that calling
S.trace('early_mean')[:] will return all samples from the last chain. One potential
problem with this is that user could simply do ``S.trace('early_mean') = 5`` and
erase the Trace object. A read-only dictionary-like class could solve
that problem.

Some backends require being closed before saving the results. This needs to be
done explicitly by the user.
"""
import pymc
import numpy as np
import types
import sys
import traceback
import warnings
import copy
__all__ = ['Trace', 'Database']

from pymc import six
from pymc import utils
from pymc.six import print_


class Trace(object):

    """Base class for Trace objects.

    Each tallyable pymc object is given a trace attribute, which is a Trace
    instance. These trace methods are generally called by the Database instance
    or by the user.
    """

    def __init__(self, name, getfunc=None, db=None):
        """Create a Trace instance.

        :Parameters:
        name : string
          The trace object name. This name should uniquely identify
          the pymc variable.
        getfunc : function
          A function returning the value to tally.
        db : Database instance
          The database owning this Trace.
        """
        self._getfunc = getfunc
        self.name = name
        self.db = db
        self._chain = -1

    def _initialize(self, chain, length):
        """Prepare for tallying. Create a new chain."""
        # If this db was loaded from the disk, it may not have its
        # tallied step methods' getfuncs yet.
        if self._getfunc is None:
            self._getfunc = self.db.model._funs_to_tally[self.name]

    def tally(self, chain):
        """Appends the object's value to a chain.

        :Parameters:
        chain : integer
          Chain index.
        """
        pass

    def truncate(self, index, chain):
        """For backends that preallocate memory, removed the unused memory."""
        pass

    def gettrace(self, burn=0, thin=1, chain=-1, slicing=None):
        """Return the trace.

        :Parameters:
          - burn (int): The number of transient steps to skip.
          - thin (int): Keep one in thin.
          - chain (int): The index of the chain to fetch. If None, return all chains.
          - slicing: A slice, overriding burn and thin assignement.
        """
        warnings.warn('Use Sampler.trace method instead.', DeprecationWarning)

    # By convention, the __call__ method is assigned to gettrace.
    __call__ = gettrace

    def __getitem__(self, i):
        """Return the trace corresponding to item (or slice) i for the chain
        defined by self._chain.
        """
        if isinstance(i, slice):
            return self.gettrace(slicing=i, chain=self._chain)
        else:
            return self.gettrace(slicing=slice(i, i + 1), chain=self._chain)

    def _finalize(self, chain):
        """Execute task necessary when tallying is over for this trace."""
        pass

    def length(self, chain=-1):
        """Return the length of the trace.

        :Parameters:
        chain : int or None
          The chain index. If None, returns the combined length of all chains.
        """
        pass

    def stats(self, alpha=0.05, start=0, batches=100,
              chain=None, quantiles=(2.5, 25, 50, 75, 97.5)):
        """
        Generate posterior statistics for node.

        :Parameters:
        name : string
          The name of the tallyable object.

        alpha : float
          The alpha level for generating posterior intervals. Defaults to
          0.05.

        start : int
          The starting index from which to summarize (each) chain. Defaults
          to zero.

        batches : int
          Batch size for calculating standard deviation for non-independent
          samples. Defaults to 100.

        chain : int
          The index for which chain to summarize. Defaults to None (all
          chains).

        quantiles : tuple or list
          The desired quantiles to be calculated. Defaults to (2.5, 25, 50, 75, 97.5).
        """

        try:
            trace = np.squeeze(
                np.array(
                    self.db.trace(
                        self.name)(
                            chain=chain),
                    float))[
                        start:]

            n = len(trace)
            if not n:
                print_(
                    'Cannot generate statistics for zero-length trace in',
                    self.__name__)
                return

            return {
                'n': n,
                'standard deviation': trace.std(0),
                'mean': trace.mean(0),
                '%s%s HPD interval' % (int(100 * (1 - alpha)), '%'): utils.hpd(trace, alpha),
                'mc error': batchsd(trace, min(n, batches)),
                'quantiles': utils.quantiles(trace, qlist=quantiles)
            }
        except:
            print_('Could not generate output statistics for', self.name)
            return


class Database(object):

    """Base Database class.

    The Database job is to create a database on disk and communicate
    with its Trace objects to tally values.
    """

    def __init__(self, dbname):
        """Create a Database instance.

        This method:
         * Assigns the local Trace class to the __Trace__ attribute.
         * Assings its name to the __name__ attribute.
         * Does whatever is necessary to set up the database.

        :Note:
        All arguments to __init__ should begin by db: dbname, dbmode, etc. This
        is critical to avoid clashes between arguments to Sampler and Database.

        This method should not be subclassed.
        """
        self.__Trace__ = Trace
        self.__name__ = 'base'
        self.dbname = dbname
        self.trace_names = []
        # A list of sequences of names of the objects to tally.
        self._traces = {}  # A dictionary of the Trace objects.
        self.chains = 0

    def _initialize(self, funs_to_tally, length=None):
        """Initialize the tallyable objects.

        Makes sure a Trace object exists for each variable and then initialize
        the Traces.

        :Parameters:
        funs_to_tally : dict
          Name- function pairs.
        length : int
          The expected length of the chain. Some database may need the argument
          to preallocate memory.
        """

        for name, fun in six.iteritems(funs_to_tally):
            if name not in self._traces:
                self._traces[
                    name] = self.__Trace__(
                        name=name,
                        getfunc=fun,
                        db=self)

            self._traces[name]._initialize(self.chains, length)

        self.trace_names.append(list(funs_to_tally.keys()))

        self.chains += 1

    def tally(self, chain=-1):
        """Append the current value of all tallyable object.

       :Parameters:
       chain : int
         The index of the chain to append the values to. By default, the values
         are appended to the last chain.
        """

        chain = range(self.chains)[chain]
        for name in self.trace_names[chain]:
            try:
                self._traces[name].tally(chain)
            except:
                cls, inst, tb = sys.exc_info()
                warnings.warn("""
Error tallying %s, will not try to tally it again this chain.
Did you make all the samevariables and step methods tallyable
as were tallyable last time you used the database file?

Error:

%s""" % (name, ''.join(traceback.format_exception(cls, inst, tb))))
                self.trace_names[chain].remove(name)

    def connect_model(self, model):
        """Link the Database to the Model instance.

        In case a new database is created from scratch, ``connect_model``
        creates Trace objects for all tallyable pymc objects defined in
        `model`.

        If the database is being loaded from an existing file, ``connect_model``
        restore the objects trace to their stored value.

        :Parameters:
        model : pymc.Model instance
          An instance holding the pymc objects defining a statistical
          model (stochastics, deterministics, data, ...)
        """
        # Changed this to allow non-Model models. -AP
        # We could also remove it altogether. -DH
        if isinstance(model, pymc.Model):
            self.model = model
        else:
            raise AttributeError('Not a Model instance.')

        # Restore the state of the Model from an existing Database.
        # The `load` method will have already created the Trace objects.
        if hasattr(self, '_state_'):
            names = set()
            for morenames in self.trace_names:
                names.update(morenames)
            for name, fun in six.iteritems(model._funs_to_tally):
                if name in self._traces:
                    self._traces[name]._getfunc = fun
                    names.discard(name)
            # if len(names) > 0:
            # print_("Some objects from the database have not been assigned a
            # getfunc", names)

        # Create a fresh new state.
        # We will be able to remove this when we deprecate traces on objects.
        else:
            for name, fun in six.iteritems(model._funs_to_tally):
                if name not in self._traces:
                    self._traces[
                        name] = self.__Trace__(
                            name=name,
                            getfunc=fun,
                            db=self)

    def _finalize(self, chain=-1):
        """Finalize the chain for all tallyable objects."""
        chain = range(self.chains)[chain]
        for name in self.trace_names[chain]:
            self._traces[name]._finalize(chain)
        self.commit()

    def truncate(self, index, chain=-1):
        """Tell the traces to truncate themselves at the given index."""
        chain = range(self.chains)[chain]
        for name in self.trace_names[chain]:
            self._traces[name].truncate(index, chain)

    def commit(self):
        """Flush data to disk."""
        pass

    def close(self):
        """Close the database."""
        self.commit()

    def savestate(self, state):
        """Store a dictionnary containing the state of the Model and its
        StepMethods."""
        self._state_ = state

    def getstate(self):
        """Return a dictionary containing the state of the Model and its
        StepMethods."""
        return getattr(self, '_state_', {})

    def trace(self, name, chain=-1):
        """Return the trace of a tallyable object stored in the database.

        :Parameters:
        name : string
          The name of the tallyable object.
        chain : int
          The trace index. Setting `chain=i` will return the trace created by
          the ith call to `sample`.
        """
        trace = copy.copy(self._traces[name])
        trace._chain = chain
        return trace


def load(dbname):
    """Return a Database instance from the traces stored on disk.

    This function should do the following:

     * Create a Database instance `db`.
     * Assign to `db` Trace instances corresponding to the pymc tallyable
       objects that are stored on disk.
     * Assign to `db` a _state_ attribute storing the stepping methods state.
       If this is not implemented, simply set db._state_ = {}.

    After loading a database `db`, we should be able to call
    the gettrace method of each tallyable object stored in the database.
    """
    pass


def batchsd(trace, batches=5):
    """
    Calculates the simulation standard error, accounting for non-independent
    samples. The trace is divided into batches, and the standard deviation of
    the batch means is calculated.
    """

    if len(np.shape(trace)) > 1:

        dims = np.shape(trace)
        # ttrace = np.transpose(np.reshape(trace, (dims[0], sum(dims[1:]))))
        ttrace = np.transpose([t.ravel() for t in trace])

        return np.reshape([batchsd(t, batches) for t in ttrace], dims[1:])

    else:
        if batches == 1:
            return np.std(trace) / np.sqrt(len(trace))

        try:
            batched_traces = np.resize(trace, (batches, int(len(trace) / batches)))
        except ValueError:
            # If batches do not divide evenly, trim excess samples
            resid = len(trace) % batches
            batched_traces = np.resize(trace[:-resid], 
                (batches, len(trace[:-resid]) / batches))

        means = np.mean(batched_traces, 1)

        return np.std(means) / np.sqrt(batches)



from pymc import *
from numpy import *
from numpy.testing import *
import nose
import sys
from pymc import utils

from pymc import six
xrange = six.moves.xrange

class test_logp_of_set(TestCase):
    A = Normal('A', 0, 1)
    B = Gamma('B', 1, 1)
    C = Lambda('C', lambda b=B: sqrt(b))
    D = Gamma('D', C, 1)

    @stochastic
    def E(x=1, value=3):
        if value != 3:
            raise RuntimeError
        else:
            return 0.
    E.value = 2

    def test_logp(self):
        self.B.rand()
        lp1 = utils.logp_of_set(set([self.A, self.B, self.D]))
        assert_almost_equal(lp1, self.A.logp + self.B.logp + self.D.logp, 10)

    def test_ZeroProb(self):
        self.B.value = -1
        for i in xrange(1000):
            try:
                utils.logp_of_set(set([self.A, self.B, self.D, self.E]))
            except:
                cls, inst, tb = sys.exc_info()
                assert(cls is ZeroProbability)

    def test_other_err(self):
        self.B.rand()
        for i in xrange(1000):
            try:
                utils.logp_of_set(set([self.A, self.B, self.D, self.E]))
            except:
                cls, inst, tb = sys.exc_info()
                assert(cls is RuntimeError)


class test_normcdf_input_shape(TestCase):

    def test_normcdf_1d_input(self):
        x = arange(8.)
        utils.normcdf(x)

    def test_normcdf_log_1d_input(self):
        x = arange(8.)
        utils.normcdf(x, log=True)

    def test_normcdf_2d_input(self):
        x = arange(8.).reshape(2, 4)
        utils.normcdf(x)

    def test_normcdf_log_2d_input(self):
        x = arange(8.).reshape(2, 4)
        utils.normcdf(x, log=True)

    def test_normcdf_3d_input(self):
        x = arange(8.).reshape(2, 2, 2)
        utils.normcdf(x)

    def test_normcdf_log_3d_input(self):
        x = arange(8.).reshape(2, 2, 2)
        utils.normcdf(x, log=True)

class test_invcdf_input_shape(TestCase):

    def test_invcdf_1d_input(self):
        x = random.random(8)
        utils.invcdf(x)
        
    def test_invcdf_2d_input(self):
        x = random.random((2, 4))
        utils.invcdf(x)

    def test_invcdf_3d_input(self):
        x = random.random((2, 2, 2))
        utils.invcdf(x)

if __name__ == '__main__':
    C = nose.config.Config(verbosity=1)
    nose.runmodule(config=C)



from pymc.gp import *
from pymc.gp.cov_funs import *
from numpy import *

# Covariance
C = Covariance(
    eval_fun=matern.euclidean,
    diff_degree=1.4,
    amp=1.,
    scale=1.)
# C = Covariance(eval_fun = pow_exp.euclidean, pow=1., amp=1., scale=1.)
# C = Covariance(eval_fun = quadratic.euclidean, phi=1., amp=1., scale=.2)
# C = Covariance(eval_fun = gaussian.euclidean, amp=1., scale=1.)
# C = Covariance(eval_fun = sphere.euclidean, amp=1., scale=.5)

# Mean


def zero_fun(x):
    return 0. * x
M = Mean(zero_fun)

# - Plot - ####
if __name__ == '__main__':
    from pylab import *

    x = arange(-1., 1., .01)

    close('all')
    figure()

    # Plot the covariance function
    subplot(2, 2, 1)

    contourf(
        x,
        x,
        C(x,
            x).view(ndarray),
        origin='lower',
        extent=(-1.,
                1.,
                -1.,
                1.),
        cmap=cm.bone)

    xlabel('x')
    ylabel('y')
    title('C(x,y)')
    axis('tight')
    colorbar()

    # Plot a slice of the covariance function
    subplot(2, 2, 2)

    plot(x, C(x, 0).view(ndarray).ravel(), 'k-')
    axis([-1, 1, 0, 1])

    xlabel('x')
    ylabel('C(x,0)')
    title('A slice of C')

    subplot(2, 1, 2)

    # plot_envelope(M, C, mesh=x)
    for i in range(3):
        f = Realization(M, C)
        plot(x, f(x))

    xlabel('x')
    ylabel('f(x)')
    title('Three realizations')
    axis([-1, 1, -2, 2])

    # show()



from numpy.testing import *
from numpy import *
from pymc.gp import Mean
from copy import copy


def quadfun(x, a, b, c):
    return (a * x ** 2 + b * x + c)

M = Mean(eval_fun=quadfun, a=1., b=.5, c=2.)


def constant(x, val):
    """docstring for parabolic_fun"""
    return zeros(x.shape[:-1], dtype=float) + val

N = Mean(constant, val=0.)

x = arange(-1, 1, .01)
y = vstack((x, x)).T


class test_mean(TestCase):

    def test(self):
        assert(M(x).shape == x.shape)
        assert(N(y).shape == (y.shape[0],))



"""Test the binary step method.
Assume we have a series of heads (0) or tail (1) experiments :
[0,0,1,1,0,0,0,1,1,0]
Assume that the coin is either fair (p=.5), or unfair (p!=.5).
We have two parameters:
 * fair : which can be True or False,
 * coin : the probability of tail on flipping the coin
"""

from numpy.testing import TestCase
import pymc
import numpy as np


class BinaryTestModel:
    series = np.array([0, 0, 1, 1, 0, 0, 0, 1, 1, 1])

    fair = pymc.Bernoulli('fair', p=.5, value=1)

    @pymc.deterministic
    def coin(fair=fair):
        if fair:
            return .5
        else:
            return .1
    # @stoch
    # def coin(value=.5, fair=fair):
    #     """Return the probability of a tail on flipping a coin."""
    #     if fair is True:
    #         return pymc.beta_like(value, 1e6, 1e6)
    #     else:
    #         return pymc.uniform_like(value, .3, .7)

    tail = pymc.Bernoulli('tail', p=coin, value=series, observed=True)


class TestBinary(TestCase):

    def test(self):
        S = pymc.MCMC(input=BinaryTestModel)
        S.sample(1000, 500, progress_bar=0)
        f = S.fair.trace()
        assert(1.0 * f.sum() / len(f) > .5)

if __name__ == '__main__':
    import unittest
    unittest.main()



# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'

import pymc as pm
from . import linalg_utils
import copy
import types
import numpy as np
from .gp_submodel import *
import warnings

from pymc import six
from pymc.six import print_
xrange = six.moves.xrange

from .Realization import Realization
from .Mean import Mean
from .Covariance import Covariance
from .GPutils import observe, regularize_array

__all__ = [
    'wrap_metropolis_for_gp_parents',
    'GPEvaluationGibbs',
    'GPParentAdaptiveMetropolis',
    'GPStepMethod',
    'GPEvaluationMetropolis',
    'MeshlessGPMetropolis']


class GPStepMethod(pm.NoStepper):

    @staticmethod
    def competence(stochastic):
        if isinstance(stochastic, GaussianProcess):
            return 1
        else:
            return 0

    def tune(self, verbose=0):
        return False

def wrap_metropolis_for_gp_parents(metro_class):
    """
    Wraps Metropolis step methods so they can handle extended parents of
    Gaussian processes.
    """
    class wrapper(metro_class):
        __doc__ = """A modified version of class %s that handles parents of Gaussian processes.
Docstring of class %s: \n\n%s""" % (metro_class.__name__, metro_class.__name__, metro_class.__doc__)

        def __init__(self, stochastic, *args, **kwds):

            self.metro_class.__init__(self, stochastic, *args, **kwds)

            mb = set(self.markov_blanket)
            for c in list(self.children):
                if isinstance(c, GaussianProcess):
                    self.children |= c.extended_children
                    mb |= c.extended_children
            self.markov_blanket = list(mb)

            # Remove f from the set that will be used to compute
            # logp_plus_loglike.
            self.markov_blanket_no_f = set(
                filter(
                    lambda x: not isinstance(
                        x,
                        GaussianProcess),
                    self.markov_blanket))
            self.fs = filter(
                lambda x: isinstance(
                    x,
                    GaussianProcess),
                self.markov_blanket)
            self.fr_checks = [f.submodel.fr_check for f in self.fs]

        def get_logp_plus_loglike(self):
            return pm.logp_of_set(self.markov_blanket_no_f)
        logp_plus_loglike = property(get_logp_plus_loglike)

        def propose(self):
            self.metro_class.propose(self)
            try:
                # First make sure none of the stochastics handled by
                # metro_method forbid their current values.
                for s in self.stochastics:
                    s.logp
                # Then make sure the covariances are all still full-rank on the
                # observation locations.
                for frc in self.fr_checks:
                    frc.logp
                for f in self.fs:
                    f.rand()
                self.f_proposed = True
            except pm.ZeroProbability:
                self.f_proposed = False

        def reject(self):
            self.metro_class.reject(self)
            if self.f_proposed:
                for f in self.fs:
                    f.revert()

        @staticmethod
        def competence(stochastic, metro_class=metro_class):
            if any([isinstance(child, GaussianProcess)
                    for child in stochastic.extended_children]):
                return metro_class.competence(stochastic) + .01
            else:
                return 0

    wrapper.__name__ = 'GPParent%s' % metro_class.__name__
    wrapper.metro_class = metro_class

    return wrapper


# Wrap all registered Metropolis step methods to use GP parents.
new_sm_dict = {}
filtered_registry = [
    sm for sm in pm.StepMethodRegistry if issubclass(
        sm,
        pm.Metropolis)]
for sm in filtered_registry:
    wrapped_method = wrap_metropolis_for_gp_parents(sm)
    new_sm_dict[wrapped_method.__name__] = wrapped_method
GPParentAdaptiveMetropolis = wrap_metropolis_for_gp_parents(
    pm.AdaptiveMetropolis)
__all__ += new_sm_dict.keys()
locals().update(new_sm_dict)


class MeshlessGPMetropolis(pm.Metropolis):

    def __init__(self, gp):
        pm.Metropolis.__init__(
            self,
            gp,
            proposal_distribution='Prior',
            check_before_accepting=False)

    def propose(self):
        self.stochastic.rand()

    @staticmethod
    def competence(stochastic):
        if isinstance(stochastic, GaussianProcess):
            if len(stochastic.submodel.mesh) == 0:
                return 3
            else:
                return 0
        else:
            return 0


class _GPEvaluationMetropolis(pm.Metropolis):

    """
    Updates a GP evaluation, the 'f_eval' attribute of a GP submodel.
    The stationary distribution of the assymetric proposal is equal
    to the prior distribution, an attempt to minimize jumps to values
    forbidden by the prior.
    """

    def __init__(self, stochastic, proposal_sd=1, **kwds):
        pm.Metropolis.__init__(
            self,
            stochastic,
            proposal_sd=proposal_sd,
            **kwds)

    def propose(self):
        sig = pm.utils.value(self.stochastic.parents['sig'])
        mu = pm.utils.value(self.stochastic.parents['mu'])

        delta = pm.rmv_normal_chol(0 * mu, sig)

        beta = np.minimum(1, self.proposal_sd * self.adaptive_scale_factor)
        bsig = beta * sig
        sb2 = np.sqrt(1 - beta ** 2)
        self.stochastic.value = (
            self.stochastic.value - mu) * sb2 + beta * delta + mu
        xp, x = self.stochastic.value, self.stochastic.last_value
        self._hastings_factor = pm.mv_normal_chol_like(
            x,
            (xp - mu) * sb2 + mu,
            bsig) - pm.mv_normal_chol_like(xp,
                                           (x - mu) * sb2 + mu,
                                           bsig)

        # self.stochastic.value = self.stochastic.value + self.adaptive_scale_factor*self.proposal_sd*delta
        # self._hastings_factor = 0

    def hastings_factor(self):
        return self._hastings_factor

    @staticmethod
    def competence(stochastic):
        if isinstance(stochastic, GPEvaluation):
            return 3
        else:
            return 0

GPEvaluationMetropolis = wrap_metropolis_for_gp_parents(
    _GPEvaluationMetropolis)


class GPEvaluationGibbs(pm.Metropolis):

    """
    Updates a GP evaluation f_eval. Assumes the only children of f_eval
    are as distributed follows:

    eps_p_f ~ Normal(f_eval[ti], 1./V)

    or

    eps_p_f ~ Normal(f_eval, 1./V)

    if ti is None.
    """

    def __init__(self, submod, V, eps_p_f, ti=None, tally=True, verbose=0):

        self.f_eval = submod.f_eval
        self.f = submod.f
        pm.StepMethod.__init__(self, [self.f, self.f_eval], tally=tally)

        self.children_no_data = copy.copy(self.children)
        if isinstance(eps_p_f, pm.Variable):
            self.children_no_data.remove(eps_p_f)
            self.eps_p_f = eps_p_f
        else:
            for epf in eps_p_f:
                self.children_no_data.remove(epf)
            self.eps_p_f = pm.Lambda(
                'eps_p_f',
                lambda e=eps_p_f: np.hstack(
                    e),
                trace=False)

        self.V = pm.Lambda(
            '%s_vect' % V.__name__,
            lambda V=V: np.resize(V,
                                  len(submod.mesh)))
        self.C_eval = submod.C_eval
        self.M_eval = submod.M_eval
        self.S_eval = submod.S_eval

        M_eval_shape = pm.utils.value(self.M_eval).shape
        C_eval_shape = pm.utils.value(self.C_eval).shape
        self.ti = ti or np.arange(M_eval_shape[0])

        # Work arrays
        self.scratch1 = np.asmatrix(np.empty(C_eval_shape, order='F'))
        self.scratch2 = np.asmatrix(np.empty(C_eval_shape, order='F'))
        self.scratch3 = np.empty(M_eval_shape)

        # Initialize hidden attributes
        self.accepted = 0.
        self.rejected = 0.
        self._state = ['rejected', 'accepted', 'proposal_distribution']
        self._tuning_info = []
        self.proposal_distribution = 'gibbs'

        self.verbose = verbose
        
    valid_proposals = property(lambda self: ('gibbs',))

    def get_logp(self):
        return 0.
    logp = property(get_logp)

    def get_loglike(self):
        return pm.utils.logp_of_set(self.children_no_data)
    loglike = property(get_loglike)

    def get_logp_plus_loglike(self):
        return self.get_loglike()
    logp_plus_loglike = property(get_logp_plus_loglike)

    def reject(self):
        self.rejected += 1
        if self.verbose:
            print_(self._id + ' rejecting')
        # Revert the field evaluation and the rest of the field.
        self.f_eval.revert()
        self.f.revert()

    def tune(self, verbose=0):
        return False

    def propose(self):

        if self.verbose:
            print_(self._id + ' proposing')

        fc = pm.gp.fast_matrix_copy

        eps_p_f = pm.utils.value(self.eps_p_f)
        f = pm.utils.value(self.f_eval)
        for i in xrange(len(self.scratch3)):
            self.scratch3[i] = np.sum(eps_p_f[self.ti[i]] - f[i])

        # Compute Cholesky factor of covariance of eps_p_f, C(x,x) + V
        C_eval_value = pm.utils.value(self.C_eval)
        C_eval_shape = C_eval_value.shape

        # Get the Cholesky factor of C_eval, plus the nugget.
        # I don't think you can use S_eval for speed, unfortunately.
        in_chol = fc(C_eval_value, self.scratch1)

        v_val = pm.utils.value(self.V)
        for i in xrange(pm.utils.value(C_eval_shape)[0]):
            in_chol[i, i] += v_val[i] / np.alen(self.ti[i])

        info = pm.gp.linalg_utils.dpotrf_wrap(in_chol)
        if info > 0:
            raise np.linalg.LinAlgError

        # Compute covariance of f conditional on eps_p_f.
        offdiag = fc(C_eval_value, self.scratch2)
        offdiag = pm.gp.trisolve(
            in_chol,
            offdiag,
            uplo='U',
            transa='T',
            inplace=True)

        C_step = offdiag.T * offdiag
        C_step *= -1
        C_step += C_eval_value

        # Compute mean of f conditional on eps_p_f.
        for i in xrange(len(self.scratch3)):
            self.scratch3[i] = np.mean(eps_p_f[self.ti[i]])
        m_step = pm.utils.value(
            self.M_eval) + np.dot(offdiag.T,
                                  pm.gp.trisolve(in_chol,
                                                 (self.scratch3 -
                                                  self.M_eval.value),
                                                 uplo='U',
                                                 transa='T')).view(np.ndarray).ravel()

        sig_step = C_step
        info = pm.gp.linalg_utils.dpotrf_wrap(C_step.T)
        if info > 0:
            warnings.warn(
                'Full conditional covariance was not positive definite.')
            return

        # Update value of f.
        self.f_eval.value = m_step + np.dot(
            sig_step,
            np.random.normal(
                size=sig_step.shape[
                    1])).view(
                        np.ndarray).ravel(
        )
        # Propose the rest of the field from its conditional prior.
        self.f.rand()



'''
Created on Jan 20, 2011

@author: jsalvatier
'''
import numpy as np
from collections import defaultdict

_sts_memory = defaultdict(dict)


def sum_to_shape(key1, key2, value, sum_shape):

    try:
        axes, lx = _sts_memory[key1][key2]

    except KeyError:

        value_shape = np.array(np.shape(value))

        sum_shape_expanded = np.zeros(value_shape.size)
        sum_shape_expanded[0:len(sum_shape)] += np.array(sum_shape)

        axes = np.where(sum_shape_expanded != value_shape)[0]
        lx = np.size(axes)

        _sts_memory[key1][key2] = (axes, lx)

    if lx > 0:
        return np.apply_over_axes(np.sum, value, axes)

    else:
        return value



"""
SQLite database backend

Store traces from tallyable objects in individual SQL tables.

Implementation Notes
--------------------
For each object, a table is created with the following format:

key (INT), trace (INT),  v1 (FLOAT), v2 (FLOAT), v3 (FLOAT) ...

For multidimensional objects, ndim>1, eg (2,2) the table has the following format:

key (INT), trace (INT),  v1_1 (FLOAT), v1_2 (FLOAT), v2_1 (FLOAT), v2_2 (FLOAT)

The key is autoincremented each time a new row is added to the table.
trace denotes the chain index, and starts at 0.

Additional Dependencies
-----------------------
sqlite3 <http://www.sqlite.org>

Changeset
---------
Created by Chris Fonnesbeck on 2007-02-01.
Updated by DH on 2007-04-04.
DB API changes, October 2008, DH.
Added support for multidimensional arrays, DH Oct. 2009
"""

# TODO: Add support for integer valued objects.

import numpy as np
from numpy import zeros, shape, squeeze, transpose
import sqlite3
from sqlite3 import OperationalError
import pymc
from pymc.database import base, pickle, ram
import pdb
import os
import warnings

__all__ = ['Trace', 'Database', 'load']


class Trace(base.Trace):

    """SQLite Trace class."""

    def _initialize(self, chain, length):
        """Create an SQL table.
        """

        if self._getfunc is None:
            self._getfunc = self.db.model._funs_to_tally[self.name]

        # Determine size
        try:
            self._shape = np.shape(self._getfunc())
        except TypeError:
            self._shape = None

        self._vstr = ', '.join(var_str(self._shape))

        # If the table already exists, exit now.
        if chain != 0:
            return

        # Create the variable name strings.
        vstr = ', '.join(v + ' FLOAT' for v in var_str(self._shape))
        query = """CREATE TABLE IF NOT EXISTS [%s]
                     (recid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
                      trace  int(5), %s)""" % (self.name, vstr)
        self.db.cur.execute(query)

    def tally(self, chain):
        """Adds current value to trace."""

        try:
            # I changed str(x) to '%f'%x to solve a bug appearing due to
            # locale settings. In french for instance, str prints a comma
            # instead of a colon to indicate the decimal, which confuses
            # the database into thinking that there are more values than there
            # is.  A better solution would be to use another delimiter than the
            # comma. -DH
            valstring = ', '.join(
                ['%f' %
                 x for x in np.ravel(self._getfunc())])
        except:
            valstring = str(self._getfunc())

        # Add value to database
        query = "INSERT INTO [%s] (recid, trace, %s) values (NULL, %s, %s)" % \
            (self.name, self._vstr, chain, valstring)
        self.db.cur.execute(query)

    def gettrace(self, burn=0, thin=1, chain=-1, slicing=None):
        """Return the trace (last by default).

        Input:
          - burn (int): The number of transient steps to skip.
          - thin (int): Keep one in thin.
          - chain (int): The index of the chain to fetch. If None, return all
            chains. By default, the last chain is returned.
          - slicing: A slice, overriding burn and thin assignement.
        """
        # warnings.warn('Use Sampler.trace method instead.',
        # DeprecationWarning)
        if not slicing:
            slicing = slice(burn, None, thin)

        # If chain is None, get the data from all chains.
        if chain is None:
            self.db.cur.execute('SELECT * FROM [%s]' % self.name)
            trace = self.db.cur.fetchall()
        else:
            # Deal with negative chains (starting from the end)
            if chain < 0:
                chain = range(self.db.chains)[chain]
            self.db.cur.execute(
                'SELECT * FROM [%s] WHERE trace=%s' %
                (self.name, chain))
            trace = self.db.cur.fetchall()
        trace = np.array(trace)[:, 2:]
        if len(self._shape) > 1:
            trace = trace.reshape(-1, *self._shape)
        return squeeze(trace[slicing])

    def __getitem__(self, index):
        chain = self._chain

        if chain is None:
            self.db.cur.execute('SELECT * FROM [%s]' % self.name)
            trace = self.db.cur.fetchall()
        else:
            # Deal with negative chains (starting from the end)
            if chain < 0:
                chain = range(self.db.chains)[chain]
            self.db.cur.execute(
                'SELECT * FROM [%s] WHERE trace=%s' %
                (self.name, chain))
            trace = self.db.cur.fetchall()

        trace = np.array(trace)[:, 2:]
        if len(self._shape) > 1:
            trace = trace.reshape(-1, *self._shape)
        else:
            trace = np.squeeze(trace)

        return trace[index]

    __call__ = gettrace

    def length(self, chain=-1):
        """Return the sample length of given chain. If chain is None,
        return the total length of all chains."""
        return len(self.gettrace(chain=chain))


class Database(base.Database):

    """SQLite database.
    """

    def __init__(self, dbname, dbmode='a'):
        """Open or create an SQL database.

        :Parameters:
        dbname : string
          The name of the database file.
        dbmode : {'a', 'w'}
          File mode.  Use `a` to append values, and `w` to overwrite
          an existing file.
        """
        self.__name__ = 'sqlite'
        self.dbname = dbname
        self.__Trace__ = Trace

        self.trace_names = []
        # A list of sequences of names of the objects to tally.
        self._traces = {}  # A dictionary of the Trace objects.

        if os.path.exists(dbname) and dbmode == 'w':
            os.remove(dbname)

        self.DB = sqlite3.connect(dbname, check_same_thread=False)
        self.cur = self.DB.cursor()

        existing_tables = get_table_list(self.cur)
        if existing_tables:
            # Get number of existing chains
            self.cur.execute(
                'SELECT MAX(trace) FROM [%s]' %
                existing_tables[0])
            self.chains = self.cur.fetchall()[0][0] + 1
            self.trace_names = self.chains * [existing_tables, ]
        else:
            self.chains = 0

    def commit(self):
        """Commit updates to database"""
        self.DB.commit()

    def close(self, *args, **kwds):
        """Close database."""
        self.cur.close()
        self.commit()
        self.DB.close()

# TODO: Code savestate and getstate to enable storing of the model's state.
# state is a dictionary with two keys: sampler and step_methods.
# state['sampler'] is another dictionary containing information about
# the sampler's state (_current_iter, _iter, _burn, etc.)
# state['step_methods'] is a dictionary with keys refering to ids for
# each samplingmethod defined in sampler.
# Each id refers to another dictionary containing the state of the
# step method.
# To do this efficiently, we would need functions that stores and retrieves
# a dictionary to and from a sqlite database. Unfortunately, I'm not familiar with
# SQL enough to do that without having to read too much SQL documentation
# for my taste.
    def savestate(self, state):
        """Store a dictionnary containing the state of the Sampler and its
        StepMethods."""
        pass

    def getstate(self):
        """Return a dictionary containing the state of the Sampler and its
        StepMethods."""
        return {}


def load(dbname):
    """Load an existing SQLite database.

    Return a Database instance.
    """
    db = Database(dbname)

    # Get the name of the objects
    tables = get_table_list(db.cur)

    # Create a Trace instance for each object
    chains = 0
    for name in tables:
        db._traces[name] = Trace(name=name, db=db)
        db._traces[name]._shape = get_shape(db.cur, name)
        setattr(db, name, db._traces[name])
        db.cur.execute('SELECT MAX(trace) FROM [%s]' % name)
        chains = max(chains, db.cur.fetchall()[0][0] + 1)

    db.chains = chains
    db.trace_names = chains * [tables, ]
    db._state_ = {}
    return db


# Copied form Django.
def get_table_list(cursor):
    """Returns a list of table names in the current database."""
    # Skip the sqlite_sequence system table used for autoincrement key
    # generation.
    cursor.execute("""
        SELECT name FROM sqlite_master
        WHERE type='table' AND NOT name='sqlite_sequence'
        ORDER BY name""")
    return [row[0] for row in cursor.fetchall()]


def get_shape(cursor, name):
    """Return the shape of the table ``name``."""
    cursor.execute('select * from [%s]' % name)
    inds = cursor.description[-1][0][1:].split('_')
    return tuple([int(i) for i in inds])


def var_str(shape):
    """Return a sequence of strings naming the element of the tallyable object.

    :Examples:
    >>> var_str((5,))
    ['v1', 'v2', 'v3', 'v4', 'v5']

    >>> var_str((2,2))
    ['v1_1', 'v1_2', 'v2_1', 'v2_2']
    """

    if shape in [None, ()]:
        return ['v1', ]

    size = np.prod(shape)
    indices = (np.indices(shape) + 1).reshape(-1, size)
    return ['v' + '_'.join(list(map(str, i))) for i in zip(*indices)]



"""
The decorators stochastic, deterministic, discrete_stochastic, binary_stochastic, potential and data
are defined here, but the actual objects are defined in PyMCObjects.py
"""

__all__ = [
    'stochastic',
    'stoch',
    'deterministic',
    'dtrm',
    'potential',
    'pot',
    'data',
    'observed',
    'robust_init',
    'disable_special_methods',
    'enable_special_methods',
    'check_special_methods']

import sys
import inspect
import pdb
from imp import load_dynamic
from .PyMCObjects import Stochastic, Deterministic, Potential
from .Node import ZeroProbability, ContainerBase, Node, StochasticMeta
from .Container import Container
import numpy as np
from .utils import get_signature

special_methods_available = [True]


def disable_special_methods(sma=special_methods_available):
    sma[0] = False


def enable_special_methods(sma=special_methods_available):
    sma[0] = True


def check_special_methods(sma=special_methods_available):
    return sma[0]

from . import six


def _extract(__func__, kwds, keys, classname, probe=True):
    """
    Used by decorators stochastic and deterministic to inspect declarations
    """

    # Add docs and name
    kwds['doc'] = __func__.__doc__
    if not 'name' in kwds:
        kwds['name'] = __func__.__name__
    # kwds.update({'doc':__func__.__doc__, 'name':__func__.__name__})

    # Instanitate dictionary of parents
    parents = {}

    # This gets used by stochastic to check for long-format logp and random:
    if probe:
        cur_status = check_special_methods()
        disable_special_methods()
        # Define global tracing function (I assume this is for debugging??)
        # No, it's to get out the logp and random functions, if they're in
        # there.

        def probeFunc(frame, event, arg):
            if event == 'return':
                locals = frame.f_locals
                kwds.update(dict((k, locals.get(k)) for k in keys))
                sys.settrace(None)
            return probeFunc

        sys.settrace(probeFunc)

        # Get the functions logp and random (complete interface).
        # Disable special methods to prevent the formation of a hurricane of
        # Deterministics
        try:
            __func__()
        except:
            if 'logp' in keys:
                kwds['logp'] = __func__
            else:
                kwds['eval'] = __func__
        # Reenable special methods.
        if cur_status:
            enable_special_methods()

    for key in keys:
        if not key in kwds:
            kwds[key] = None

    for key in ['logp', 'eval']:
        if key in keys:
            if kwds[key] is None:
                kwds[key] = __func__

    # Build parents dictionary by parsing the __func__tion's arguments.
    (args, defaults) = get_signature(__func__)

    if defaults is None:
        defaults = ()

    # Make sure all parents were defined
    arg_deficit = (len(args) - ('value' in args)) - len(defaults)
    if arg_deficit > 0:
        err_str =  classname + ' ' + __func__.__name__ + \
            ': no parent provided for the following labels:'
        for i in range(arg_deficit):
            err_str += " " + args[i + ('value' in args)]
            if i < arg_deficit - 1:
                err_str += ','
        raise ValueError(err_str)

    # Fill in parent dictionary
    try:
        parents.update(dict(zip(args[-len(defaults):], defaults)))
    except TypeError:
        pass

    value = parents.pop('value', None)

    return (value, parents)


def stochastic(__func__=None, __class__=Stochastic,
               binary=False, discrete=False, **kwds):
    """
    Decorator function for instantiating stochastic variables. Usages:

    Medium:

        @stochastic
        def A(value = ., parent_name = .,  ...):
            return foo(value, parent_name, ...)

        @stochastic(trace=trace_object)
        def A(value = ., parent_name = .,  ...):
            return foo(value, parent_name, ...)

    Long:

        @stochastic
        def A(value = ., parent_name = .,  ...):

            def logp(value, parent_name, ...):
                return foo(value, parent_name, ...)

            def random(parent_name, ...):
                return bar(parent_name, ...)


        @stochastic(trace=trace_object)
        def A(value = ., parent_name = .,  ...):

            def logp(value, parent_name, ...):
                return foo(value, parent_name, ...)

            def random(parent_name, ...):
                return bar(parent_name, ...)

    where foo() computes the log-probability of the variable A
    conditional on its value and its parents' values, and bar()
    generates a random value from A's distribution conditional on
    its parents' values.

    :SeeAlso:
      Stochastic, Deterministic, deterministic, data, Potential, potential, Model,
      distributions
    """

    def instantiate_p(__func__):
        value, parents = _extract(__func__, kwds, keys, 'Stochastic')
        return __class__(value=value, parents=parents, **kwds)

    keys = ['logp', 'random', 'rseed']

    instantiate_p.kwds = kwds

    if __func__:
        return instantiate_p(__func__)

    return instantiate_p

# Shortcut alias
stoch = stochastic


def potential(__func__=None, **kwds):
    """
    Decorator function instantiating potentials. Usage:

    @potential
    def B(parent_name = ., ...)
        return baz(parent_name, ...)

    where baz returns the deterministic B's value conditional
    on its parents.

    :SeeAlso:
      Deterministic, deterministic, Stochastic, Potential, stochastic, data, Model
    """
    def instantiate_pot(__func__):
        junk, parents = _extract(
            __func__, kwds, keys, 'Potential', probe=False)
        return Potential(parents=parents, **kwds)

    keys = ['logp']

    instantiate_pot.kwds = kwds

    if __func__:
        return instantiate_pot(__func__)

    return instantiate_pot
pot = potential


def deterministic(__func__=None, **kwds):
    """
    Decorator function instantiating deterministic variables. Usage:

    @deterministic
    def B(parent_name = ., ...)
        return baz(parent_name, ...)

    @deterministic(trace = trace_object)
    def B(parent_name = ., ...)
        return baz(parent_name, ...)

    where baz returns the variable B's value conditional
    on its parents.

    :SeeAlso:
      Deterministic, Potential, potential, Stochastic, stochastic, data, Model,
      CommonDeterministics
    """
    def instantiate_n(__func__):
        junk, parents = _extract(
            __func__, kwds, keys, 'Deterministic', probe=False)
        return Deterministic(parents=parents, **kwds)

    keys = ['eval']

    instantiate_n.kwds = kwds

    if __func__:
        return instantiate_n(__func__)

    return instantiate_n

# Shortcut alias
dtrm = deterministic


def observed(obj=None, **kwds):
    """
    Decorator function to instantiate data objects.
    If given a Stochastic, sets a the observed flag to True.

    Can be used as

    @observed
    def A(value = ., parent_name = .,  ...):
        return foo(value, parent_name, ...)

    or as

    @stochastic(observed=True)
    def A(value = ., parent_name = .,  ...):
        return foo(value, parent_name, ...)


    :SeeAlso:
      stochastic, Stochastic, dtrm, Deterministic, potential, Potential, Model,
      distributions
    """

    if obj is not None:
        if isinstance(obj, Stochastic):
            obj._observed = True
            return obj
        else:
            p = stochastic(__func__=obj, observed=True, **kwds)
            return p

    kwds['observed'] = True

    def instantiate_observed(func):
        return stochastic(func, **kwds)

    return instantiate_observed

data = observed


def robust_init(stochclass, tries, *args, **kwds):
    """Robust initialization of a Stochastic.

    If the evaluation of the log-probability returns a ZeroProbability
    error, due for example to a parent being outside of the support for
    this Stochastic, the values of parents are randomly sampled until
    a valid log-probability is obtained.

    If the log-probability is still not valid after `tries` attempts, the
    original ZeroProbability error is raised.

    :Parameters:
    stochclass : Stochastic, eg. Normal, Uniform, ...
      The Stochastic distribution to instantiate.
    tries : int
      Maximum number of times parents will be sampled.
    *args, **kwds
      Positional and keyword arguments to declare the Stochastic variable.

    :Example:
    >>> lower = pymc.Uniform('lower', 0., 2., value=1.5, rseed=True)
    >>> pymc.robust_init(pymc.Uniform, 100, 'data', lower=lower, upper=5, value=[1,2,3,4], observed=True)
    """
    # Find the direct parents
    stochs = [arg for arg in (list(args) + list(kwds.values()))
              if isinstance(arg.__class__, StochasticMeta)]

    # Find the extended parents
    parents = stochs
    for s in stochs:
        parents.extend(s.extended_parents)

    extended_parents = set(parents)

    # Select the parents with a random method.
    random_parents = [
        p for p in extended_parents if p.rseed is True and hasattr(
            p,
            'random')]

    for i in range(tries):
        try:
            return stochclass(*args, **kwds)
        except ZeroProbability:
            exc = sys.exc_info()
            for parent in random_parents:
                try:
                    parent.random()
                except:
                    six.reraise(*exc)

    six.reraise(*exc)



#
#
# Test of convergence diagnostics
#
#

from __future__ import with_statement
from numpy.testing import assert_equal, assert_array_equal
from numpy.testing import assert_approx_equal, TestCase
import unittest
import nose
import numpy as np
import pymc
import pymc.examples.weibull_fit as model
import os
import warnings
import copy

np.random.seed(467)

iterations = 10000
burnin = 9000

S = pymc.MCMC(model, 'ram')
S.sample(iterations, burnin, progress_bar=0)

# Known data for testing integrated autocorrelation time = 2.28
x = np.array([0.98073604, 0.98073604, 0.98073604, 0.98073604, 0.98073604,
              0.41424798, 0.58398493, 0.27391045, 0.27391045, 0.27391045,
              0.27391045, 0.27391045, 0.72886149, 0.72886149, 0.72886149,
              0.67478139, 0.67478139, 0.67478139, 0.67478139, 0.67478139,
              0.67478139, 0.27720909, 0.6026456, 0.6026456, 0.47108579,
              0.47108579, 0.47108579, 0.47108579, 0.47108579, 0.47108579,
              0.47108579, 0.47108579, 0.47108579, 0.47108579, 0.47108579,
              0.47108579, 0.47108579, 0.47108579, 0.47108579, 0.47108579,
              0.47108579, 0.47108579, 0.47108579, 0.47108579, 0.47108579,
              0.34546653, 0.34546653, 0.5441314, 0.5441314, 0.5441314,
              0.5441314, 0.5441314, 0.5441314, 0.5441314, 0.5441314,
              0.37344506, 0.37344506, 0.83126209, 0.83126209, 0.3439339,
              0.3439339, 0.3439339, 0.34551721, 0.34551721, 0.34551721,
              0.44112754, 0.44112754, 0.44112754, 0.55397635, 0.55397635,
              0.55397635, 0.55397635, 0.55397635, 0.55397635, 0.55397635,
              0.55397635, 0.55397635, 0.55397635, 0.55397635, 0.55397635,
              0.55397635, 0.55397635, 0.55397635, 0.55397635, 0.55397635,
              0.55397635, 0.36521137, 0.36521137, 0.36521137, 0.36521137,
              0.36521137, 0.36521137, 0.36521137, 0.36521137, 0.36521137,
              0.36521137, 0.36521137, 0.36521137, 0.36521137, 0.32755692])


DIR = 'testresults'


class test_geweke(TestCase):
    
    try:
        import statsmodels
    except ImportError:
        raise nose.SkipTest

    def test_independent(self):
        # Use IID data

        x = [pymc.geweke(np.random.normal(size=1000),intervals=5, maxlag=5)[0][1] 
            for _ in range(1000)]


        assert_approx_equal(np.var(x), 1, 1)

        # If the model has converged, 95% the scores should lie
        # within 2 standard deviations of zero, under standard normal model
        intervals = 40
        x = np.transpose(pymc.geweke(
                np.random.normal(size=10000),intervals=intervals))[1]
        assert(sum(np.abs(x) < 2) >= int(0.9 * intervals))

    def test_simple(self):

        intervals = 20
        
        scores = pymc.geweke(S, intervals=intervals, maxlag=5)
        a_scores = scores['a']
        assert_equal(len(a_scores), intervals)

        # Plot diagnostics (if plotting is available)
        try:
            from pymc.Matplot import geweke_plot as plot
            plot(scores, path=DIR)
        except ImportError:
            pass


class test_effective_n(TestCase):
    """Unit test for effective sample size"""
    
    def test_independent_normal(self, m=3, n=1000, k=1000):
        
        n_eff = np.mean([pymc.effective_n(np.random.normal(size=(m,n))) for i in range(1000)])
        
        assert_approx_equal(n_eff, m*n, 2)
        


class test_gelman_rubin(TestCase):

    """Unit test for Gelman-Rubin diagnostic"""

    def test_fail(self):

        pass

    def test_simple(self):

        S2 = copy.copy(S)
        S2.sample(iterations, burnin, progress_bar=0)

        gr = pymc.gelman_rubin(S2)

        for i in gr:
            assert_approx_equal(gr[i], 1., 2)


class test_iat(TestCase):

    def test_simple(self):

        iat = pymc.iat(x)

        # IAT should be approximately 2.28
        assert_approx_equal(iat, 2.28, 2)

if __name__ == "__main__":

    original_filters = warnings.filters[:]
    warnings.simplefilter("ignore")
    try:
        import nose
        C = nose.config.Config(verbosity=1)
        nose.runmodule(config=C)
    finally:
        warnings.filters = original_filters

    # TODO: Restore in 2.2
    # with warnings.catch_warnings():
    #         warnings.simplefilter('ignore', FutureWarning)
    #         import nose
    #         C =nose.config.Config(verbosity=1)
    #         nose.runmodule(config=C)
